"""Keyboard shortcut definitions and utils."""

import enum
from collections.abc import Callable

from PySide6.QtCore import QObject
from PySide6.QtGui import QKeySequence, QShortcut


class Shortcut(enum.StrEnum):
    """Helper for setting keyboard shortcuts."""

    def connect(self, parent: QObject, func: Callable[[], None]) -> None:
        QShortcut(QKeySequence(self.value), parent).activated.connect(func)


class MainWindowShortcut(Shortcut):
    """Keyboard shortcuts within the main window."""

    SELECT_FOLDER = "Ctrl+O"
    PAUSE_DOWNLOAD = "Ctrl+Shift+Return"
    OPEN_PREFERENCES = "Ctrl+,"
    OPEN_DEBUG_CONSOLE = "Ctrl+."
    GO_TO_SEARCH = "Ctrl+F"
    GO_TO_SONG_TABLE = "Ctrl+T"
    GO_TO_FILTERS = "Ctrl+S"
    GO_TO_FILTER_SEARCH = "Ctrl+Shift+F"


class SongTableShortcut(Shortcut):
    """Keyboard shortcuts usable when the song table is focused."""

    DOWNLOAD = "Ctrl+Return"
    ABORT_DOWNLOAD = "Ctrl+Alt+Return"
    TRASH_SONG = "Ctrl+D"
    PIN_SONG = "Ctrl+P"
    PLAY_SAMPLE = "Space"
    PREVIEW = "Shift+Space"
    OPEN_SONG = "Ctrl+Shift+O"


class DebugConsoleShortcut(Shortcut):
    """Keyboard shortcuts within the debug console window."""

    EXECUTE = "Ctrl+Return"
    EXECUTE_PRINT = "Ctrl+Shift+Return"
    CLEAR_OUTPUT = "Ctrl+L"
    CLEAR_INPUT = "Ctrl+Shift+L"



from __future__ import annotations

import base64
import functools
import itertools
import logging
import socket
from io import BytesIO
from typing import Any, ClassVar

import flask
import qrcode
import werkzeug.serving
from PySide6 import QtCore
from werkzeug.serving import WSGIRequestHandler

from usdb_syncer import SongId, db, logger, utils
from usdb_syncer.usdb_song import UsdbSong


def _get_songs(request: flask.Request) -> list[UsdbSong]:
    builder = db.SearchBuilder()
    builder.statuses = [db.DownloadStatus.OUTDATED, db.DownloadStatus.SYNCHRONIZED]
    if search := request.args.get("search"):
        builder.text = search
    match request.args.get("sort_by", "artist"):
        case "artist":
            builder.order = db.SongOrder.ARTIST
        case "title":
            builder.order = db.SongOrder.TITLE
        case "year":
            builder.order = db.SongOrder.YEAR
    sort_order = request.args.get("sort_order", "asc")
    if sort_order == "desc":
        builder.descending = True
    limit = request.args.get("limit", 100, type=int)
    offset = request.args.get("offset", 0, type=int)

    song_ids = db.search_usdb_songs(builder)
    songs = [
        s
        for i in itertools.islice(song_ids, offset, offset + limit)
        if (s := UsdbSong.get(i))
    ]

    return songs


def _index(title: str) -> str:
    songs = _get_songs(flask.request)
    search = flask.request.args.get("search", default="")
    sort_by = flask.request.args.get("sort_by", "artist")
    sort_order = flask.request.args.get("sort_order", "asc")
    offset = flask.request.args.get("offset", 0, type=int)
    if "HX-Request" in flask.request.headers:
        return flask.render_template(
            "songs_table.html",
            songs=songs,
            search=search,
            sort_by=sort_by,
            sort_order=sort_order,
            offset=offset,
        )
    return flask.render_template(
        "index.html",
        title=title,
        qrcode=base64.b64encode(get_qrcode(address())).decode("utf-8"),
        songs=songs,
        search=search,
        sort_by=sort_by,
        sort_order=sort_order,
        offset=offset,
    )


def _api_songs() -> str:
    songs = _get_songs(flask.request)
    search = flask.request.args.get("search", default="")
    sort_by = flask.request.args.get("sort_by", "artist")
    sort_order = flask.request.args.get("sort_order", "asc")
    offset = flask.request.args.get("offset", 0, type=int)
    return flask.render_template(
        "songs_rows.html",
        songs=songs,
        search=search,
        sort_by=sort_by,
        sort_order=sort_order,
        offset=offset,
    )


def _api_mp3() -> flask.Response:
    song_id = flask.request.args.get("song_id", type=int)
    if not song_id:
        return flask.abort(400, "song_id parameter is required")
    song = UsdbSong.get(SongId(song_id))
    if not song:
        return flask.abort(404, "Song not found")
    audio_path = song.audio_path()
    if not audio_path or not audio_path.is_file():
        return flask.abort(404, "Audio file not found for this song")
    return flask.send_file(audio_path, mimetype="audio/mp3")


def _create_app(title: str) -> flask.Flask:
    app = flask.Flask(__name__)

    @app.route("/")
    def index() -> str:
        return _index(title)

    @app.route("/api/songs")
    def api_songs() -> str:
        return _api_songs()

    @app.route("/api/mp3")
    def api_mp3() -> flask.Response:
        return _api_mp3()

    @app.before_request
    def before_request() -> None:
        db.connect(utils.AppPaths.db)

    @app.teardown_request
    def teardown_request(exception: Any = None) -> None:
        db.close()

    return app


class _CustomRequestHandler(WSGIRequestHandler):
    def log(self, type: str, message: str, *args: Any) -> None:  # noqa: A002
        # don't log requests to info level
        super().log("debug", message, *args)


class _WebserverThread(QtCore.QThread):
    def __init__(self, server: werkzeug.serving.BaseWSGIServer) -> None:
        super().__init__()
        self.server = server

    def run(self) -> None:
        self.server.serve_forever()


DEFAULT_TITLE = "USDB Syncer Song Collection"


class _WebserverManager:
    _server: ClassVar[werkzeug.serving.BaseWSGIServer | None] = None
    _thread: ClassVar[QtCore.QThread | None] = None
    host: ClassVar[str] = ""
    port: ClassVar[int] = 0
    title: ClassVar[str] = DEFAULT_TITLE

    @classmethod
    def start(
        cls, host: str | None = None, port: int | None = None, title: str | None = None
    ) -> None:
        if cls._server:
            return
        if title:
            cls.title = title
        app = _create_app(cls.title)
        cls.host = host or get_local_ip()
        cls._server = werkzeug.serving.make_server(
            cls.host, port or 0, app, request_handler=_CustomRequestHandler
        )
        cls.port = cls._server.socket.getsockname()[1]
        cls._thread = _WebserverThread(cls._server)
        cls._thread.start()

    @classmethod
    def stop(cls) -> None:
        if cls._server:
            cls._server.shutdown()
            cls._server = None
        if cls._thread:
            cls._thread.quit()
            cls._thread.wait()
            cls._thread = None

    @classmethod
    def is_running(cls) -> bool:
        return bool(cls._server)


def start(
    host: str | None = None, port: int | None = None, title: str | None = None
) -> None:
    logging.getLogger("werkzeug").setLevel(logging.DEBUG)
    _WebserverManager.start(host=host, port=port, title=title)
    logger.logger.info(f"Webserver is now running on {address()}")


def stop() -> None:
    _WebserverManager.stop()
    logger.logger.info("Webserver was stopped.")


def is_running() -> bool:
    return _WebserverManager.is_running()


def address() -> str:
    return f"http://{_WebserverManager.host}:{_WebserverManager.port}"


@functools.lru_cache(maxsize=1)
def get_qrcode(data: str) -> bytes:
    buf = BytesIO()
    qrcode.make(data, box_size=5, border=2).save(buf)
    return buf.getvalue()


def get_local_ip() -> str:
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.settimeout(5)
    try:
        sock.connect(("10.255.255.255", 1))
        ip = sock.getsockname()[0]
    except OSError:
        ip = "127.0.0.1"
    finally:
        sock.close()
    return ip



"""Dialog to preview a downloaded song."""

from __future__ import annotations

import functools
import subprocess
from collections.abc import Callable, Generator
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, NewType, TypeVar

import attrs
import numpy as np
import soundfile
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import Qt

from usdb_syncer import SongId, utils
from usdb_syncer.gui import events, icons, theme
from usdb_syncer.gui.forms.PreviewWindow import Ui_MainWindow
from usdb_syncer.gui.resources.audio import METRONOME_TICK_WAV
from usdb_syncer.logger import logger as _logger
from usdb_syncer.logger import song_logger
from usdb_syncer.song_txt import SongTxt, tracks
from usdb_syncer.song_txt.auxiliaries import BeatsPerMinute
from usdb_syncer.usdb_song import UsdbSong

if TYPE_CHECKING:
    import sounddevice
else:
    try:
        import sounddevice
    except (OSError, ImportError):
        _logger.exception(
            "Failed to load module 'sounddevice', some features may be unavailable."
        )
        sounddevice = None

Seconds = NewType("Seconds", float)
Milliseconds = NewType("Milliseconds", int)
Sample = NewType("Sample", int)
Pixel = NewType("Pixel", int)
Fraction = NewType("Fraction", float)
"""A relative size, i.e. from the interval [0, 1]."""

_PITCH_ROWS = 16
_EPS_SECS = Seconds(0.01)
_INT16_MAX = 2**15 - 1
_INT16_MIN = -(2**15)
_SAMPLE_RATE = Sample(44100)
_SYNTH_AMPLITUDE = 0.4
_SYNTH_ATTACK_SECS = Seconds(0.01)
_SYNTH_RELEASE_SECS = Seconds(0.05)
_SYNTH_ATTACK_SAMPLES = Sample(int(_SYNTH_ATTACK_SECS * _SAMPLE_RATE))
_SYNTH_RELEASE_SAMPLES = Sample(int(_SYNTH_RELEASE_SECS * _SAMPLE_RATE))

_MISSING_LIB_HELP = (
    "https://github.com/bohning/usdb_syncer?tab=readme-ov-file#linux-compatibility"
)
_MISSING_LIB_MSG = (
    f"This feature is unavailable due to missing libraries, see {_MISSING_LIB_HELP}."
)
_NO_DEVICES_MSG = (
    "No audio devices found. This likely means your platform is not supported for this"
    " feature."
)


T = TypeVar("T", int, float)


def clamp(value: T, lower: T, upper: T) -> T:
    return min(max(value, lower), upper)


class Previewer(Ui_MainWindow, QtWidgets.QMainWindow):
    """Dialog to preview a downloaded song."""

    _instance: ClassVar[Previewer | None] = None
    _MIN_COVER_SIZE = Pixel(100)
    _MAX_COVER_SIZE = Pixel(200)
    _FPS = 60
    _REFRESH_RATE = Milliseconds(1000 // _FPS)
    _DOUBLECLICK_DELAY_MS = Milliseconds(200)
    _DOUBLECLICK_DELAY_SECS = Seconds(_DOUBLECLICK_DELAY_MS / 1000)

    def __init__(
        self, song_id: SongId | None, txt: SongTxt, audio: Path, cover: Path | None
    ) -> None:
        super().__init__()
        self.song_id = song_id
        self.setupUi(self)
        self._insert_cover_widget()
        self.setWindowTitle(txt.headers.artist_title_str())
        self._setup_voice_selection(txt)
        self._state = _PlayState.new(txt, audio)
        self._player = _AudioPlayer.new(audio, self._state)
        self._setup_views()
        self._set_song_info(txt, cover)
        self._connect_ui_inputs()
        self._setup_timers()
        self._on_theme_changed(theme.Theme.from_settings())
        events.ThemeChanged.subscribe(lambda e: self._on_theme_changed(e.theme))

    @classmethod
    def load(
        cls,
        *,
        txt: SongTxt,
        audio_path: Path,
        cover_path: Path | None,
        song_id: SongId | None,
    ) -> bool:
        if not sounddevice:
            QtWidgets.QMessageBox.warning(None, "Aborted", _MISSING_LIB_MSG)
            return False
        if not sounddevice.query_devices():
            QtWidgets.QMessageBox.warning(None, "Aborted", _NO_DEVICES_MSG)
            return False
        if cls._instance:
            cls._instance._change_song(song_id, txt, audio_path, cover_path)
            cls._instance.raise_()
        else:
            cls._instance = cls(song_id, txt, audio_path, cover_path)
            cls._instance.show()
        return True

    @classmethod
    def load_song(cls, song: UsdbSong) -> bool:
        if not (txt_path := song.txt_path()) or not (
            txt := SongTxt.try_from_file(txt_path, song_logger(song.song_id))
        ):
            QtWidgets.QMessageBox.warning(None, "Aborted", "Txt file is invalid!")
            return False
        if not (audio_path := song.audio_path()):
            QtWidgets.QMessageBox.warning(
                None, "Aborted", "Song must have txt and audio files!"
            )
            return False
        return cls.load(
            txt=txt,
            audio_path=audio_path,
            cover_path=song.cover_path(),
            song_id=song.song_id,
        )

    @classmethod
    def load_txt(cls, txt_path: Path) -> bool:
        if not (txt := SongTxt.try_from_file(txt_path, _logger)):
            QtWidgets.QMessageBox.warning(None, "Aborted", "Txt file is invalid!")
            return False
        if not (audio_file := txt.headers.audio or txt.headers.mp3):
            QtWidgets.QMessageBox.warning(
                None, "Aborted", "Song txt does not define an audio source!"
            )
            return False
        if not (audio_path := txt_path.parent / audio_file).is_file():
            QtWidgets.QMessageBox.warning(
                None, "Aborted", f"Audio source '{audio_path}' not found!"
            )
            return False
        if (
            not txt.headers.cover
            or not (cover_path := txt_path.parent / txt.headers.cover).is_file()
        ):
            cover_path = None
        return cls.load(
            txt=txt, audio_path=audio_path, cover_path=cover_path, song_id=None
        )

    @classmethod
    def close_song(cls, song_id: SongId) -> None:
        if cls._instance and cls._instance.song_id == song_id:
            cls._instance.close()

    def _change_song(
        self, song_id: SongId | None, txt: SongTxt, audio: Path, cover: Path | None
    ) -> None:
        self.song_id = song_id
        self._player.stop()
        self.setWindowTitle(txt.headers.artist_title_str())
        self._setup_voice_selection(txt)
        self._state.change_song(txt, audio)
        self._player.change_source(audio)
        self._set_song_info(txt, cover)

    def _setup_voice_selection(self, txt: SongTxt) -> None:
        self.label_voice.setVisible(bool(txt.notes.track_2))
        self.comboBox_voice.clear()
        self.comboBox_voice.setVisible(bool(txt.notes.track_2))
        self.comboBox_voice.addItem(txt.headers.p1 or "P1")
        self.comboBox_voice.addItem(txt.headers.p2 or "P2")

    def _setup_views(self) -> None:
        palette = theme.Theme.from_settings().preview_palette()
        self._song_view = _SongView(self._state, palette, self._on_drag)
        self.layout_main.insertWidget(0, self._song_view, 10)
        self._line_view = _LineView(self._state, palette)
        self.layout_main.insertWidget(1, self._line_view, 10)
        self.layout_main.insertStretch(2, 1)

    def _connect_ui_inputs(self) -> None:
        self.button_pause.toggled.connect(self._on_pause_toggled)
        self.button_to_start.pressed.connect(self._on_seek_start)
        self.button_backward.pressed.connect(self._on_seek_backward)
        self.button_forward.pressed.connect(self._on_seek_forward)
        self.button_to_end.pressed.connect(self._on_seek_end)
        self._on_source_volume_changed(self.slider_source.value())
        self._on_ticks_volume_changed(self.slider_ticks.value())
        self._on_pitch_volume_changed(self.slider_pitch.value())
        self.slider_source.valueChanged.connect(self._on_source_volume_changed)
        self.slider_ticks.valueChanged.connect(self._on_ticks_volume_changed)
        self.slider_pitch.valueChanged.connect(self._on_pitch_volume_changed)
        self.comboBox_voice.currentIndexChanged.connect(self._state.set_voice)

    def _set_song_info(self, txt: SongTxt, cover: Path | None) -> None:
        self._cover_label.setPixmap(QtGui.QPixmap(cover) if cover else QtGui.QPixmap())
        self.label_bpm.setText(f"#BPM: {txt.headers.bpm}")
        self.label_gap.setText(f"#GAP: {txt.headers.gap}")
        self.label_start.setText(f"#START: {txt.headers.start or '-'}")
        self.label_end.setText(f"#END: {txt.headers.end or '-'}")

    def _insert_cover_widget(self) -> None:
        self._cover_label = _SquareLabel("")
        self._cover_label.setScaledContents(True)
        self._cover_label.setMinimumSize(self._MIN_COVER_SIZE, self._MIN_COVER_SIZE)
        self._cover_label.setMaximumSize(self._MAX_COVER_SIZE, self._MAX_COVER_SIZE)
        self.layout_extra.insertWidget(1, self._cover_label)

    def _setup_timers(self) -> None:
        self._update_timer = QtCore.QTimer(self, interval=self._REFRESH_RATE)
        self._update_timer.timeout.connect(self._update_time)
        self._update_timer.start()
        self._seek_timer = QtCore.QTimer(
            self, singleShot=True, interval=self._DOUBLECLICK_DELAY_MS
        )
        self._seek_timer.timeout.connect(self._on_seek_timeout)

    def _update_time(self) -> None:
        # only get current time from playback stream if not currently seeking
        if not self._state.seeking:
            self._state.calculate_video_time_from_audio()
        self._song_view.update()
        self._line_view.update()
        current = _secs_to_str(self._state.current_video_time)
        song = _secs_to_str(self._state.song_duration)
        self.label_time.setText(f"{current} / {song}")

    def _on_pause_toggled(self, paused: bool) -> None:
        self._state.paused = paused

    def _on_seek_start(self) -> None:
        self._start_seeking(-self._state.current_video_line_idx)

    def _on_seek_end(self) -> None:
        self._start_seeking(
            len(self._state.lines) - self._state.current_video_line_idx - 1
        )

    def _on_seek_backward(self) -> None:
        line_elapsed = (
            self._state.current_video_time - self._state.current_video_line.start
        )
        if line_elapsed > self._DOUBLECLICK_DELAY_SECS:
            self._start_seeking(0)
        elif self._state.current_video_line_idx > 0:
            self._start_seeking(-1)

    def _on_seek_forward(self) -> None:
        if (
            self._state.current_video_line_idx == 0
            and self._state.current_video_time + _EPS_SECS
            < self._state.current_video_line.start
        ):
            self._start_seeking(0)
        elif self._state.current_video_line_idx < len(self._state.lines) - 1:
            self._start_seeking(1)

    def _start_seeking(self, line_delta: int) -> None:
        self._state.seeking = True
        idx = self._state.current_video_line_idx + line_delta
        self._state.set_current_video_time(self._state.lines[idx].start)
        self._seek_timer.start()

    def _on_drag(self, target_time: Seconds, released: bool) -> None:
        self._state.seeking = not released
        self._seek_timer.stop()
        self._state.set_current_video_time(Seconds(target_time))
        if released:
            self._player.seek_to(self._state.current_video_time)

    def _on_seek_timeout(self) -> None:
        self._player.seek_to(self._state.current_video_time)
        self._state.seeking = False

    def _on_theme_changed(self, theme: theme.Theme) -> None:
        self.button_pause.setIcon(icons.Icon.PAUSE_REMOTE.icon(theme.KEY))
        self.button_to_start.setIcon(icons.Icon.SKIP_TO_START.icon(theme.KEY))
        self.button_backward.setIcon(icons.Icon.SKIP_BACKWARD.icon(theme.KEY))
        self.button_forward.setIcon(icons.Icon.SKIP_FORWARD.icon(theme.KEY))
        self.button_to_end.setIcon(icons.Icon.SKIP_TO_END.icon(theme.KEY))
        self._song_view.colors = self._line_view.colors = theme.preview_palette()

    def _on_source_volume_changed(self, value: int) -> None:
        # ticks are in 10% steps
        value = value * 10
        self.label_source_volume.setText(str(value))
        self._state.source_volume = _percentage_to_amplitude_factor(value)

    def _on_ticks_volume_changed(self, value: int) -> None:
        value = value * 10
        self.label_ticks_volume.setText(str(value))
        self._state.ticks_volume = _percentage_to_amplitude_factor(value)

    def _on_pitch_volume_changed(self, value: int) -> None:
        value = value * 10
        self.label_pitch_volume.setText(str(value))
        self._state.pitch_volume = _percentage_to_amplitude_factor(value)

    def _on_voice_changed(self, voice: int) -> None:
        pass

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:  # noqa: N802
        self._player.stop()
        Previewer._instance = None
        event.accept()


def _secs_to_str(seconds: Seconds) -> str:
    secs = int(seconds)
    return f"{secs // 60:02}:{secs % 60:02}"


class _SquareLabel(QtWidgets.QLabel):
    """Custom widget to contain the cover image and maintain sqaured."""

    def hasHeightForWidth(self) -> bool:  # noqa: N802
        return True

    def heightForWidth(self, width: int) -> int:  # noqa: N802
        return width

    def sizeHint(self) -> QtCore.QSize:  # noqa: N802
        base = super().sizeHint()
        side = min(base.width(), base.height())
        return QtCore.QSize(side, side)


@attrs.define
class _PlayState:
    """State of the current playback."""

    # shared state
    _tracks: list[list[_Line]]
    lines: list[_Line] = attrs.field(init=False)
    song_duration: Seconds
    seeking: bool = False
    paused: bool = False

    # video state
    current_video_line_idx: int = 0
    current_video_line: _Line = attrs.field(init=False)
    current_video_time: Seconds = attrs.field(init=False)
    current_pos_in_line: Fraction = Fraction(0.0)

    # audio state
    current_sample: Sample = Sample(0)
    _audio_note_iter: Generator[_Note, None, None] = attrs.field(init=False)
    current_audio_note: _Note | None = attrs.field(init=False)
    source_volume: float = 0.0
    ticks_volume: float = 0.0
    pitch_volume: float = 0.0

    @classmethod
    def new(cls, txt: SongTxt, audio: Path) -> _PlayState:
        song_duration = Seconds(utils.get_media_duration(audio))
        track_list = list(filter(None, [txt.notes.track_1, txt.notes.track_2]))
        tracks_ = [
            [
                _Line.new(
                    line, Milliseconds(txt.headers.gap), txt.headers.bpm, song_duration
                )
                for line in track
            ]
            for track in track_list
        ]
        state = cls(tracks=tracks_, song_duration=song_duration)
        state._set_initial_video_time(txt)
        return state

    def _set_initial_video_time(self, txt: SongTxt) -> None:
        if txt.headers.previewstart is not None:
            start = Seconds(txt.headers.previewstart)
        else:
            start = self.lines[0].start
        self.set_current_video_time(start)

    def __attrs_post_init__(self) -> None:
        self.lines = self._tracks[0]
        self.current_video_line = self.lines[0]
        self.current_video_time = self.lines[0].start
        self._audio_note_iter = self._iter_notes()
        self.current_audio_note = next(self._audio_note_iter, None)

    def change_song(self, txt: SongTxt, audio: Path) -> None:
        new = _PlayState.new(txt, audio)
        self._tracks = new._tracks
        self.song_duration = new.song_duration
        self.__attrs_post_init__()
        self._set_initial_video_time(txt)

    def calculate_video_time_from_audio(self) -> None:
        self.set_current_video_time(Seconds(self.current_sample / _SAMPLE_RATE))

    def set_current_video_time(self, secs: Seconds) -> None:
        self.current_video_time = secs
        self.current_pos_in_line = Fraction(
            self.current_video_time / self.song_duration
        )
        self.current_video_line_idx, self.current_video_line = next(
            (i, li)
            for i, li in enumerate(self.lines)
            if li.line_break is None or li.line_break > self.current_video_time
        )

    def advance_current_sample(self, delta: int) -> None:
        self.current_sample = Sample(self.current_sample + delta)
        self._advance_to_next_unplayed_note()

    def set_current_sample_from_secs(self, secs: float) -> None:
        self.current_sample = Sample(int(secs * _SAMPLE_RATE))
        self._reset_current_audio_note()

    def _reset_current_audio_note(self) -> None:
        self._audio_note_iter = self._iter_notes()
        self.current_audio_note = next(self._audio_note_iter, None)
        self._advance_to_next_unplayed_note()

    def _advance_to_next_unplayed_note(self) -> None:
        while (
            self.current_audio_note
            and self.current_sample > self.current_audio_note.end_sample
        ):
            self.current_audio_note = next(self._audio_note_iter, None)

    def _iter_notes(self) -> Generator[_Note, None, None]:
        for line in self.lines:
            yield from line.notes

    def set_voice(self, voice: int) -> None:
        self.lines = self._tracks[voice]
        self.set_current_video_time(self.current_video_time)
        self._reset_current_audio_note()


def _beat_to_secs(beat: int, gap: Milliseconds, bpm: BeatsPerMinute) -> Seconds:
    return Seconds(gap / 1000 + bpm.beats_to_secs(beat))


def _beat_to_sample(beat: int, gap: Milliseconds, bpm: BeatsPerMinute) -> Sample:
    return Sample(int(_beat_to_secs(beat, gap, bpm) * _SAMPLE_RATE))


@attrs.define
class _Line:
    """Information for rendering a line."""

    notes: list[_Note]
    start: Seconds
    end: Seconds
    duration: Seconds
    line_break: Seconds | None
    rel_x_pos_in_song: Fraction
    rel_width_in_song: Fraction
    text: str

    @classmethod
    def new(
        cls,
        line: tracks.Line,
        gap: Milliseconds,
        bpm: BeatsPerMinute,
        song_duration: Seconds,
    ) -> _Line:
        start = _beat_to_secs(line.start(), gap, bpm)
        end = _beat_to_secs(line.end(), gap, bpm)
        duration = Seconds(end - start)
        if line.line_break:
            line_break = _beat_to_secs(line.line_break.previous_line_out_time, gap, bpm)
        else:
            line_break = None
        return cls(
            notes=_Note.from_line(line, gap, bpm),
            start=start,
            end=end,
            duration=duration,
            rel_x_pos_in_song=Fraction(start / song_duration),
            rel_width_in_song=Fraction(duration / song_duration),
            line_break=line_break,
            text="".join(n.text for n in line.notes).strip(),
        )


@attrs.define
class _Note:
    """Information for rendering and playing a note."""

    text: str
    x_pos: Fraction
    width: Fraction
    y_pos: Fraction

    freq: float
    start_sample: Sample
    end_sample: Sample
    kind: tracks.NoteKind

    @classmethod
    def new(
        cls,
        note: tracks.Note,
        line_start: int,
        line_len: int,
        line_pitch: range,
        gap: Milliseconds,
        bpm: BeatsPerMinute,
    ) -> _Note:
        x_pos = Fraction((note.start - line_start) / line_len)
        width = Fraction(note.duration / line_len)
        # move to the middle of available space
        shift = max(_PITCH_ROWS - len(line_pitch), 0) // 2
        y_pos = Fraction(1 - (note.pitch - line_pitch.start + shift) / _PITCH_ROWS)
        return cls(
            text=note.text,
            kind=note.kind,
            x_pos=x_pos,
            width=width,
            y_pos=y_pos,
            freq=_ultrastar_pitch_to_freq(note.pitch),
            start_sample=_beat_to_sample(note.start, gap, bpm),
            end_sample=_beat_to_sample(note.end(), gap, bpm),
        )

    @classmethod
    def from_line(
        cls, line: tracks.Line, gap: Milliseconds, bpm: BeatsPerMinute
    ) -> list[_Note]:
        start = line.start()
        duration = line.end() - start
        min_pitch = min(n.pitch for n in line.notes)
        max_pitch = max(n.pitch for n in line.notes)
        pitch = range(min_pitch, max_pitch + 1)
        return [_Note.new(n, start, duration, pitch, gap, bpm) for n in line.notes]


class _SongView(QtWidgets.QWidget):
    _is_dragging = False
    _MIN_HEIGHT = Pixel(20)
    _MAX_HEIGHT = Pixel(40)
    _NEEDLE_WIDTH = Pixel(2)

    def __init__(
        self,
        state: _PlayState,
        colors: theme.PreviewPalette,
        on_drag: Callable[[Seconds, bool], None],
    ):
        super().__init__()
        self._state = state
        self.colors = colors
        self._on_drag = on_drag
        self.setMinimumHeight(self._MIN_HEIGHT)
        self.setMaximumHeight(self._MAX_HEIGHT)
        self.setCursor(Qt.CursorShape.OpenHandCursor)

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:  # noqa: N802
        height = self.height()
        width = self.width()
        with QtGui.QPainter(self) as painter:
            painter.setPen(Qt.PenStyle.NoPen)
            painter.setBrush(self.colors.line)
            for line in self._state.lines:
                x = round(line.rel_x_pos_in_song * width)
                w = round(line.rel_width_in_song * width)
                painter.drawRect(x, 0, w, height)

            painter.setPen(QtGui.QPen(self.colors.needle, self._NEEDLE_WIDTH))
            x_pos = round(self._state.current_pos_in_line * width)
            half_needle = self._NEEDLE_WIDTH // 2
            x_pos = clamp(x_pos, half_needle, width - half_needle)
            painter.drawLine(x_pos, 0, x_pos, height)

    def _current_time_at_mouse(self, event: QtGui.QMouseEvent) -> Seconds:
        return Seconds(
            self._state.song_duration * clamp(event.pos().x() / self.width(), 0.0, 1.0)
        )

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:  # noqa: N802
        if event.button() == Qt.MouseButton.LeftButton:
            self._is_dragging = True
            self.setCursor(Qt.CursorShape.ClosedHandCursor)
            self._on_drag(self._current_time_at_mouse(event), False)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:  # noqa: N802
        if self._is_dragging:
            self._on_drag(self._current_time_at_mouse(event), False)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:  # noqa: N802
        if self._is_dragging:
            self._is_dragging = False
            self.setCursor(Qt.CursorShape.OpenHandCursor)
            self._on_drag(self._current_time_at_mouse(event), True)


@attrs.define
class _LinePaintContext:
    """Values for painting a line."""

    total_width: Pixel
    row_height: Pixel
    notes_height: Pixel
    current_pos: Fraction
    radius: float
    text_height: Pixel

    _RELATIVE_TEXT_ROW_HEIGHT = 2

    @classmethod
    def new(cls, view: _LineView) -> _LinePaintContext:
        total_height = Pixel(view.height())
        total_width = Pixel(view.width())

        # divide into equally sized pitch rows plus one scaled row for text
        row_height = Pixel(
            round(total_height / (_PITCH_ROWS + cls._RELATIVE_TEXT_ROW_HEIGHT))
        )
        notes_height = Pixel(row_height * _PITCH_ROWS)
        text_height = Pixel(total_height - notes_height)

        line_elapsed = (
            view._state.current_video_time - view._state.current_video_line.start
        )
        current_pos = Fraction(line_elapsed / view._state.current_video_line.duration)

        return cls(
            total_width=total_width,
            row_height=row_height,
            notes_height=notes_height,
            current_pos=current_pos,
            # make sides circular
            radius=row_height / 2,
            text_height=text_height,
        )


class _LineView(QtWidgets.QWidget):
    _POINTER_SIZE_IN_ROWS = 1.5
    _GRID_LINE_WIDTH = Pixel(1)
    _MIN_SIZE = (Pixel(600), Pixel(200))
    _MAX_HEIGHT = Pixel(600)

    def __init__(self, state: _PlayState, colors: theme.PreviewPalette):
        super().__init__()
        self._state = state
        self.colors = colors
        self.setMinimumSize(*self._MIN_SIZE)
        self.setMaximumHeight(self._MAX_HEIGHT)

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:  # noqa: N802
        with QtGui.QPainter(self) as painter:
            ctx = _LinePaintContext.new(self)
            self._draw_grid(painter, ctx)
            self._draw_notes(painter, ctx)
            self._draw_text(painter, ctx)
            self._draw_pointer(painter, ctx)

    def _draw_grid(self, painter: QtGui.QPainter, ctx: _LinePaintContext) -> None:
        painter.setPen(QtGui.QPen(self.colors.grid, self._GRID_LINE_WIDTH))
        for i in range(1, _PITCH_ROWS):
            y = ctx.notes_height - i * ctx.row_height
            painter.drawLine(0, y, ctx.total_width, y)

    def _draw_notes(self, painter: QtGui.QPainter, ctx: _LinePaintContext) -> None:
        painter.setPen(Qt.PenStyle.NoPen)
        for note in self._state.current_video_line.notes:
            painter.setBrush(self._brush_for_note(note, ctx))
            x = round(note.x_pos * ctx.total_width)
            net_row_height = ctx.row_height - self._GRID_LINE_WIDTH
            y = round(note.y_pos * ctx.notes_height) - net_row_height
            w = round(note.width * ctx.total_width)
            painter.drawRoundedRect(x, y, w, net_row_height, ctx.radius, ctx.radius)

    def _brush_for_note(self, note: _Note, ctx: _LinePaintContext) -> QtGui.QBrush:
        if ctx.current_pos > note.x_pos:
            if note.kind.is_golden():
                color = self.colors.active_golden_note
            else:
                color = self.colors.active_note
        else:
            if note.kind.is_golden():
                color = self.colors.golden_note
            else:
                color = self.colors.note
        if note.kind.has_pitch():
            style = Qt.BrushStyle.SolidPattern
        else:
            style = Qt.BrushStyle.Dense5Pattern
        return QtGui.QBrush(color, style)

    def _draw_text(self, painter: QtGui.QPainter, ctx: _LinePaintContext) -> None:
        font = painter.font()
        font.setPixelSize(ctx.text_height * 2 // 3)
        font_metrics = QtGui.QFontMetrics(font)
        # for simplicity's sake, don't take account for italicization
        text_width = font_metrics.horizontalAdvance(self._state.current_video_line.text)
        text_start = (ctx.total_width - text_width) // 2
        font.setItalic(True)
        italic_font_metrics = QtGui.QFontMetrics(font)
        for note in self._state.current_video_line.notes:
            font.setItalic(not note.kind.has_pitch())
            painter.setFont(font)
            active = ctx.current_pos > note.x_pos
            painter.setPen(self.colors.active_text if active else self.colors.text)
            painter.drawText(
                QtCore.QRect(text_start, ctx.notes_height, text_width, ctx.text_height),
                Qt.AlignmentFlag.AlignVCenter,
                note.text,
            )
            metrics = font_metrics if note.kind.has_pitch() else italic_font_metrics
            text_start += metrics.horizontalAdvance(note.text)

    def _draw_pointer(self, painter: QtGui.QPainter, ctx: _LinePaintContext) -> None:
        pointer_size = round(ctx.row_height * self._POINTER_SIZE_IN_ROWS)
        x = round(ctx.current_pos * ctx.total_width)
        current_pitch = next(
            (
                n
                for n in reversed(self._state.current_video_line.notes)
                if n.x_pos < ctx.current_pos
            ),
            self._state.current_video_line.notes[0],
        ).y_pos
        if current_pitch > 0.1:
            y_point = round(current_pitch * ctx.notes_height) - ctx.row_height
            y_base = y_point - pointer_size
        else:
            # close to top border, show pointer below note
            # also ensure it's in view, even if the note is too high
            y_point = max(round(current_pitch * ctx.notes_height), 0)
            y_base = y_point + pointer_size
        points = (
            QtCore.QPoint(x, y_point),
            QtCore.QPoint(x - pointer_size // 2, y_base),
            QtCore.QPoint(x + pointer_size // 2, y_base),
        )
        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(self.colors.active_note)
        painter.drawPolygon(points)


@attrs.define
class _AudioPlayer:
    _source: Path
    _state: _PlayState
    _tick_data: np.ndarray
    _stream: sounddevice.OutputStream | None = None
    _process: subprocess.Popen | None = None
    _CHANNELS = 2
    _STREAM_BUFFER_SIZE = Sample(2048)

    @classmethod
    def new(cls, source: Path, state: _PlayState) -> _AudioPlayer:
        data, _samplerate = soundfile.read(METRONOME_TICK_WAV, dtype="float32")
        player = cls(source=source, state=state, tick_data=data)
        player._start_ffmpeg(state.current_video_time)
        player._start_stream()
        return player

    def change_source(self, source: Path) -> None:
        self.stop()
        self._source = source
        self._start_ffmpeg(self._state.current_video_time)
        self._start_stream()

    def stop(self) -> None:
        if self._stream:
            self._stream.stop()
            self._stream.close()
        if self._process:
            self._process.terminate()
            self._process = None

    def seek_to(self, seconds: Seconds) -> None:
        self.stop()
        self._start_ffmpeg(seconds)
        self._start_stream()

    def _start_ffmpeg(self, start: Seconds = Seconds(0.0)) -> None:
        self._state.set_current_sample_from_secs(start)
        cmd = [
            "ffmpeg",
            "-ss",
            str(start),
            "-i",
            str(self._source),
            "-f",
            "s16le",
            "-acodec",
            "pcm_s16le",
            "-ar",
            str(_SAMPLE_RATE),
            "-ac",
            str(self._CHANNELS),
            "-",
        ]
        self._process = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
        )

    def _start_stream(self) -> None:
        self._stream = sounddevice.OutputStream(
            samplerate=_SAMPLE_RATE,
            channels=self._CHANNELS,
            dtype="int16",
            blocksize=self._STREAM_BUFFER_SIZE,
            callback=self._stream_callback,
        )
        self._stream.start()

    def _stream_callback(
        self,
        outdata: np.ndarray,
        samples: Sample,
        time: Any,
        status: sounddevice.CallbackFlags,
    ) -> None:
        if (
            not (self._process and self._process.stdout and self._stream)
            or self._state.paused
        ):
            outdata[:] = np.zeros((samples, self._CHANNELS), dtype=np.int16)
            return
        bytes_needed = samples * self._CHANNELS * 2
        data = self._process.stdout.read(bytes_needed)
        if len(data) < bytes_needed:
            outdata[:] = np.zeros((samples, self._CHANNELS), dtype=np.int16)
            self._stream.stop()
            return
        array = np.frombuffer(data, dtype=np.int16).reshape(-1, self._CHANNELS)
        outdata[:] = self._mix_audio(array, samples)
        self._state.advance_current_sample(samples)

    def _mix_audio(self, data: np.ndarray, samples: Sample) -> np.ndarray:
        ticks = self._get_tick_start_and_data(samples)
        pitch = self._get_pitch_start_and_data(samples)
        if ticks is None and pitch is None and self._state.source_volume == 1.0:
            # nothing to mix; skip conversion
            return data
        data = data.astype(np.float32) / _INT16_MAX
        data *= self._state.source_volume
        if ticks is not None:
            tick_start, tick_data = ticks
            data[tick_start : tick_start + len(tick_data)] += tick_data
        if pitch is not None:
            pitch_start, pitch_data = pitch
            data[pitch_start : pitch_start + len(pitch_data)] += pitch_data
        return np.clip(data * _INT16_MAX, _INT16_MIN, _INT16_MAX).astype(np.int16)

    def _get_tick_start_and_data(
        self, samples: Sample
    ) -> tuple[Sample, np.ndarray] | None:
        if (note := self._state.current_audio_note) is None:
            return None
        if self._state.ticks_volume == 0.0:
            return None

        start_sample = self._state.current_sample
        end_sample = start_sample + samples
        tick_end = note.start_sample + len(self._tick_data)
        overlap_start = max(start_sample, note.start_sample)
        overlap_end = min(end_sample, tick_end)
        if overlap_start >= overlap_end:
            return None
        main_offset = Sample(overlap_start - start_sample)
        tick_offset = overlap_start - note.start_sample
        length = overlap_end - overlap_start

        data = self._tick_data[tick_offset : tick_offset + length]
        return main_offset, data * self._state.ticks_volume

    def _get_pitch_start_and_data(
        self, samples: Sample
    ) -> tuple[Sample, np.ndarray] | None:
        if (note := self._state.current_audio_note) is None:
            return None
        if self._state.pitch_volume == 0.0:
            return None
        if not note.kind.has_pitch():
            return None

        start_sample = self._state.current_sample
        end_sample = start_sample + samples
        overlap_start = max(start_sample, note.start_sample)
        overlap_end = min(end_sample, note.end_sample)
        if overlap_start >= overlap_end:
            return None
        main_offset = Sample(overlap_start - start_sample)
        pitch_offset = overlap_start - note.start_sample
        length = overlap_end - overlap_start

        data = _synth_note(
            note.freq, note.end_sample - note.start_sample, self._CHANNELS
        )[pitch_offset : pitch_offset + length]
        return main_offset, data * self._state.pitch_volume


def _percentage_to_amplitude_factor(value: int) -> float:
    """Given a percentage of volume, return the amplitude factor.

    - 0%   -> silence
    - 100% -> original
    - 200% -> double perceived loudness, i.e. +10dB
    """
    return 10 ** ((value / 100 - 1) / 2) if value > 0 else 0


def _ultrastar_to_midi_pitch(pitch: int) -> float:
    return pitch + 60


def _ultrastar_pitch_to_freq(pitch: int) -> float:
    # https://en.wikipedia.org/wiki/MIDI_tuning_standard
    return 440.0 * 2 ** ((_ultrastar_to_midi_pitch(pitch) - 69) / 12.0)


@functools.lru_cache(1)
def _synth_note(freq: float, samples: Sample, channels: int) -> np.ndarray:
    t = np.linspace(0, samples / _SAMPLE_RATE, samples, endpoint=False)
    waveform = _SYNTH_AMPLITUDE * np.sin(2 * np.pi * freq * t)
    envelope = _get_synth_envelope(samples)
    mono_wave = (waveform * envelope).astype(np.float32)
    return np.column_stack([mono_wave for _ in range(channels)])


def _get_synth_envelope(samples: Sample) -> np.ndarray:
    attack_samples = min(_SYNTH_ATTACK_SAMPLES, samples)
    release_samples = min(_SYNTH_RELEASE_SAMPLES, samples - attack_samples)
    sustain_samples = samples - (attack_samples + release_samples)
    attack = np.linspace(0, 1, attack_samples)
    sustain = np.ones(sustain_samples)
    release = np.linspace(1, 0, release_samples)
    return np.concatenate((attack, sustain, release))



"""High-level routines for USDB and local songs."""

from __future__ import annotations

import json
import os
from collections.abc import Generator
from importlib import resources
from importlib.resources.abc import Traversable
from pathlib import Path

import attrs
import requests
from requests import Session

from usdb_syncer import (
    SongId,
    SyncMetaId,
    data,
    db,
    errors,
    events,
    settings,
    song_txt,
    utils,
)
from usdb_syncer.logger import error_logger, logger
from usdb_syncer.song_loader import DownloadManager
from usdb_syncer.sync_meta import SyncMeta
from usdb_syncer.usdb_scraper import get_updated_songs_from_usdb
from usdb_syncer.usdb_song import UsdbSong, UsdbSongEncoder
from usdb_syncer.utils import AppPaths


def load_available_songs_and_sync_meta(folder: Path, force_reload: bool) -> None:
    """Load available songs from USDB and synchronize the sync meta folder."""
    with db.transaction():
        result = load_available_songs(force_reload=force_reload)
        synchronize_sync_meta_folder(folder, not result.synced_with_usdb)
        SyncMeta.reset_active(folder)
    UsdbSong.clear_cache()
    initialize_auto_downloads(result.new_songs)


@attrs.define
class LoadSongsResult:
    """Result of loading songs from cache and USDB."""

    new_songs: set[SongId] = attrs.field(factory=set)
    synced_with_usdb: bool = False


def load_available_songs(
    force_reload: bool, session: Session | None = None
) -> LoadSongsResult:
    """True if USDB was queried successfully."""
    result = LoadSongsResult()
    if force_reload:
        last = db.LastUsdbUpdate.zero()
        UsdbSong.delete_all()
    elif (last := db.LastUsdbUpdate.get()).is_zero() and (songs := load_cached_songs()):
        UsdbSong.upsert_many(songs)
        result.new_songs.update((s.song_id for s in songs))
        last = db.LastUsdbUpdate.get()
    try:
        songs = get_updated_songs_from_usdb(last, session=session)
    except errors.UsdbLoginError:
        logger.debug("Skipping fetching new songs as there is no login.")
    except requests.exceptions.ConnectionError:
        logger.debug("", exc_info=True)
        logger.error("Failed to fetch new songs; check network connection.")
    else:
        result.synced_with_usdb = True
        if songs:
            UsdbSong.upsert_many(songs)
            result.new_songs.update((s.song_id for s in songs))
    return result


def initialize_auto_downloads(updates: set[SongId]) -> None:
    if not utils.ffmpeg_is_available():
        return
    download_ids = set(db.SavedSearch.get_subscribed_song_ids()).intersection(updates)
    if settings.get_auto_update():
        search = db.SearchBuilder(statuses=[db.DownloadStatus.OUTDATED])
        download_ids.update(db.search_usdb_songs(search))
    songs = [s for i in download_ids if (s := UsdbSong.get(i))]
    if not songs:
        return
    for song in songs:
        song.status = db.DownloadStatus.PENDING
    with db.transaction():
        UsdbSong.upsert_many(songs)
    events.DownloadsRequested(len(songs)).post()
    DownloadManager.download(songs)


def load_cached_songs() -> list[UsdbSong] | None:
    if AppPaths.song_list.exists():
        path: Path | Traversable = AppPaths.song_list
    elif (resource := resources.files(data).joinpath("song_list.json")).is_file():
        path = resource
    else:
        return None
    with path.open(encoding="utf8") as file:
        try:
            return json.load(file, object_hook=UsdbSong.from_json)
        except (json.decoder.JSONDecodeError, TypeError, KeyError):
            return None


def dump_available_songs(songs: list[UsdbSong], target: Path | None = None) -> None:
    with (target or AppPaths.song_list).open("w", encoding="utf8") as file:
        json.dump(songs, file, cls=UsdbSongEncoder)


def _iterate_usdb_files_in_folder_recursively(
    folder: Path,
) -> Generator[Path, None, None]:
    for root, _, files in os.walk(folder):
        for file in files:
            if file.endswith(".usdb") and not file.startswith("."):
                yield Path(root) / file


@attrs.define
class _SyncMetaFolderSyncer:
    folder: Path
    keep_unknown_song_ids: bool
    db_metas: dict[SyncMetaId, SyncMeta]
    song_ids: set[SongId]
    to_upsert: list[SyncMeta]
    found_metas: set[SyncMetaId]

    @classmethod
    def new(cls, folder: Path, keep_unknown_song_ids: bool) -> _SyncMetaFolderSyncer:
        return cls(
            folder=folder,
            keep_unknown_song_ids=keep_unknown_song_ids,
            db_metas={m.sync_meta_id: m for m in SyncMeta.get_in_folder(folder)},
            song_ids=set(db.all_song_ids()),
            to_upsert=[],
            found_metas=set(),
        )

    def process(self) -> None:
        for path in _iterate_usdb_files_in_folder_recursively(folder=self.folder):
            self._process_path(path)
        SyncMeta.delete_many_in_folder(
            self.folder, tuple(self.db_metas.keys() - self.found_metas)
        )
        SyncMeta.upsert_many(self.to_upsert)

    def _process_path(self, path: Path) -> None:
        if meta_id := SyncMetaId.from_path(path):
            if meta_id in self.found_metas:
                utils.trash_or_delete_path(path)
                logger.warning(f"Trashed duplicated meta file: '{path}'")
                return
            self.found_metas.add(meta_id)

        meta = None if meta_id is None else self.db_metas.get(meta_id)

        if meta_id is not None and meta and meta.mtime == utils.get_mtime(path):
            self._process_unchanged_file(path, meta)
        else:
            self._process_changed_or_new_file(path)

    def _process_unchanged_file(self, path: Path, meta: SyncMeta) -> None:
        if not utils.compare_unicode_paths(path, meta.path):
            meta.path = path
            self.to_upsert.append(meta)
            logger.info(f"Meta file was moved: '{path}'.")

    def _process_changed_or_new_file(self, path: Path) -> None:
        if not (meta := SyncMeta.try_from_file(path)):
            return
        if meta.song_id in self.song_ids:
            # file was changed and maybe moved
            self.to_upsert.append(meta)
            if meta.sync_meta_id in self.db_metas:
                logger.info(f"Updated meta file from disk: '{path}'.")
            else:
                logger.info(f"New meta file found on disk: '{path}'.")
        elif not self.keep_unknown_song_ids:
            logger.info(f"{meta.song_id.usdb_detail_url()} no longer exists on USDB.")
            if settings.get_trash_remotely_deleted_songs():
                logger.info(f"Deleting '{path.parent}' locally as well.")
                utils.trash_or_delete_path(path.parent)


def synchronize_sync_meta_folder(folder: Path, keep_unknown_song_ids: bool) -> None:
    _SyncMetaFolderSyncer.new(folder, keep_unknown_song_ids).process()


def find_local_songs(directory: Path) -> set[SongId]:
    matched_rows: set[SongId] = set()
    for path in directory.glob("**/*.txt"):
        if headers := try_parse_txt_headers(path):
            name = headers.artist_title_str()
            if matches := list(
                db.find_similar_usdb_songs(
                    utils.normalize(headers.artist), utils.normalize(headers.title)
                )
            ):
                plural = "es" if len(matches) > 1 else ""
                logger.info(f"{len(matches)} match{plural} for '{name}'.")
                matched_rows.update(matches)
            else:
                logger.warning(f"No matches for '{name}'.")
    return matched_rows


def try_parse_txt_headers(path: Path) -> song_txt.Headers | None:
    if lines := utils.read_file_head(path, 20):
        try:
            return song_txt.Headers.parse(lines, error_logger)
        except errors.HeadersParseError:
            return None
    return None



"""Representations of rows of the filter tree."""

from __future__ import annotations

import enum
from collections.abc import Iterable
from typing import Generic, TypeVar, assert_never

import attrs
from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon

from usdb_syncer import db
from usdb_syncer.custom_data import CustomData
from usdb_syncer.gui.icons import Icon


# not an ABC so enums can inherit it (otherwise there's a metaclass conflict)
class TreeItemData:
    """Interface for objects to be rendered as a tree node."""

    def decoration(self) -> QIcon | None:
        raise NotImplementedError

    def is_checkable(self, has_checked_children: bool) -> bool:
        raise NotImplementedError

    def is_editable(self) -> bool:
        raise NotImplementedError

    def is_parent(self) -> bool:
        raise NotImplementedError

    def is_accepted(self, _matches: set[str | int]) -> bool:
        raise NotImplementedError

    def build_search(self, search: db.SearchBuilder) -> None:
        raise NotImplementedError

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        raise NotImplementedError

    def child_data(self) -> Iterable[TreeItemData]:
        raise NotImplementedError


class RootItemData(TreeItemData):
    """Implementation of the tree root."""

    def decoration(self) -> QIcon | None:
        return None

    def is_checkable(self, has_checked_children: bool) -> bool:
        return False

    def is_editable(self) -> bool:
        return False

    def is_parent(self) -> bool:
        return True

    def is_accepted(self, _matches: set[str | int]) -> bool:
        return True

    def build_search(self, search: db.SearchBuilder) -> None:
        pass

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return True

    def child_data(self) -> Iterable[TreeItemData]:
        return Filter


@attrs.define(kw_only=True)
class TreeItem:
    """A row in the tree."""

    data: TreeItemData = attrs.field(factory=RootItemData)
    parent: TreeItem | None = None
    row_in_parent: int = attrs.field(default=0, init=False)
    children: list[TreeItem] = attrs.field(factory=list, init=False)
    checked: bool | None = attrs.field(default=False, init=False)
    checked_children: set[int] = attrs.field(factory=set, init=False)
    flags: Qt.ItemFlag = attrs.field(default=Qt.ItemFlag.ItemIsEnabled, init=False)

    def __attrs_post_init__(self) -> None:
        if self.data.is_checkable(True):
            self.flags |= Qt.ItemFlag.ItemIsUserCheckable
        else:
            self.checked = None
        if self.data.is_editable():
            self.flags |= Qt.ItemFlag.ItemIsEditable
        if not self.data.is_parent():
            self.flags |= Qt.ItemFlag.ItemNeverHasChildren

    def populate(self) -> None:
        self.set_children(TreeItem(data=d, parent=self) for d in self.data.child_data())
        for child in self.children:
            child.populate()

    def set_children(self, children: Iterable[TreeItem]) -> None:
        self.children = list(children)
        for row, child in enumerate(self.children):
            child.parent = self
            child.row_in_parent = row

    def add_child(self, child: TreeItem) -> None:
        child.parent = self
        child.row_in_parent = len(self.children)
        self.children.append(child)

    def remove_child(self, child: TreeItem) -> None:
        self.children.remove(child)
        self.checked_children.discard(child.row_in_parent)
        if self.checked is not None:
            self.checked = bool(self.checked_children)
        for later_child in self.children[child.row_in_parent :]:
            if later_child.row_in_parent in self.checked_children:
                self.checked_children.remove(later_child.row_in_parent)
                self.checked_children.add(later_child.row_in_parent - 1)
            later_child.row_in_parent -= 1

    def toggle_checked(self, keep_siblings: bool) -> list[TreeItem]:
        if not self.is_checkable():
            return []
        if self.checked:
            changed: list[TreeItem] = self._uncheck_children()
        else:
            changed = [self]
            self.checked = True
        if self.parent and self.parent.data.is_checkable(True):
            changed += self.parent._update_after_child_toggled(self, keep_siblings)
        return changed

    def is_checkable(self) -> bool:
        return self.data.is_checkable(bool(self.checked_children))

    def _uncheck_children(self) -> list[TreeItem]:
        changed: list[TreeItem] = [self]
        for child in self._checked_child_items():
            changed += child._uncheck_children()
        self.checked_children.clear()
        self.checked = False
        return changed

    def _update_after_child_toggled(
        self, child: TreeItem, keep_siblings: bool
    ) -> list[TreeItem]:
        checked_before = self.checked
        changed: list[TreeItem] = []
        if child.checked:
            if not keep_siblings:
                # `child` is not yet in children, so this is fine
                changed = self._uncheck_children()
            self.checked_children.add(child.row_in_parent)
        else:
            self.checked_children.remove(child.row_in_parent)
        self.checked = bool(self.checked_children)
        if self.checked != checked_before:
            changed.append(self)
            if self.parent and self.parent.data.is_checkable(True):
                changed += self.parent._update_after_child_toggled(
                    self, keep_siblings=True
                )
        return changed

    def apply_search(self, search: db.SearchBuilder) -> list[TreeItem]:
        changed = []
        if self.children:
            for child in self.children:
                before = child.checked
                changed.extend(child.apply_search(search))
                if self.checked is not None and child.checked != before:
                    if child.checked:
                        self.checked_children.add(child.row_in_parent)
                    else:
                        self.checked_children.remove(child.row_in_parent)
            if self.checked is not None and self.checked != bool(self.checked_children):
                changed.append(self)
                self.checked = not self.checked
        elif (
            self.checked is not None and self.data.is_in_search(search) != self.checked
        ):
            self.checked = not self.checked
            changed.append(self)
        return changed

    def _checked_child_items(self) -> Iterable[TreeItem]:
        return (self.children[row] for row in self.checked_children)

    def build_search(self, search: db.SearchBuilder) -> None:
        if self.checked is None:
            for child in self.children:
                child.build_search(search)
        elif self.checked:
            self.data.build_search(search)
            for child in self._checked_child_items():
                child.build_search(search)

    def filter_accepts_child(
        self, child_row: int, filters: dict[Filter, set[str | int]]
    ) -> bool:
        if (
            not isinstance(self.data, Filter)
            or (matches := filters.get(self.data)) is None
        ):
            return True
        return self.children[child_row].data.is_accepted(matches)


class Filter(TreeItemData, enum.Enum):
    """Kinds of filters in the tree."""

    SAVED = 0
    STATUS = enum.auto()
    ARTIST = enum.auto()
    TITLE = enum.auto()
    EDITION = enum.auto()
    LANGUAGE = enum.auto()
    GOLDEN_NOTES = enum.auto()
    RATING = enum.auto()
    VIEWS = enum.auto()
    YEAR = enum.auto()
    GENRE = enum.auto()
    CREATOR = enum.auto()
    CUSTOM_DATA = enum.auto()

    def __str__(self) -> str:  # noqa: C901
        match self:
            case Filter.SAVED:
                return "Saved Searches"
            case Filter.STATUS:
                return "Status"
            case Filter.ARTIST:
                return "Artist"
            case Filter.TITLE:
                return "Title"
            case Filter.EDITION:
                return "Edition"
            case Filter.LANGUAGE:
                return "Language"
            case Filter.GOLDEN_NOTES:
                return "Golden Notes"
            case Filter.RATING:
                return "Rating"
            case Filter.VIEWS:
                return "Views"
            case Filter.YEAR:
                return "Year"
            case Filter.GENRE:
                return "Genre"
            case Filter.CREATOR:
                return "Creator"
            case Filter.CUSTOM_DATA:
                return "Custom Data"
            case _ as unreachable:
                assert_never(unreachable)

    def child_data(self) -> Iterable[TreeItemData]:  # noqa: C901
        match self:
            case Filter.SAVED:
                return SavedSearch.load_all()
            case Filter.ARTIST:
                return (SongArtistMatch(v, c) for v, c in db.usdb_song_artists())
            case Filter.TITLE:
                return (SongTitleMatch(v, c) for v, c in db.usdb_song_titles())
            case Filter.EDITION:
                return (SongEditionMatch(v, c) for v, c in db.usdb_song_editions())
            case Filter.LANGUAGE:
                return (SongLanguageMatch(v, c) for v, c in db.usdb_song_languages())
            case Filter.STATUS:
                return StatusVariant
            case Filter.GOLDEN_NOTES:
                return GoldenNotesVariant
            case Filter.RATING:
                return RatingVariant
            case Filter.VIEWS:
                return ViewsVariant
            case Filter.YEAR:
                return (SongYearMatch(v, c) for v, c in db.usdb_song_years())
            case Filter.GENRE:
                return (SongGenreMatch(v, c) for v, c in db.usdb_song_genres())
            case Filter.CREATOR:
                return (SongCreatorMatch(v, c) for v, c in db.usdb_song_creators())
            case Filter.CUSTOM_DATA:
                return (CustomDataKeyMatch(k) for k in sorted(CustomData.key_options()))
            case _ as unreachable:
                assert_never(unreachable)

    def decoration(self) -> QIcon:  # noqa: C901
        match self:
            case Filter.SAVED:
                icon = Icon.SAVED_SEARCH
            case Filter.STATUS:
                icon = Icon.DOWNLOAD
            case Filter.ARTIST:
                icon = Icon.ARTIST
            case Filter.TITLE:
                icon = Icon.TITLE
            case Filter.EDITION:
                icon = Icon.EDITION
            case Filter.LANGUAGE:
                icon = Icon.LANGUAGE
            case Filter.GOLDEN_NOTES:
                icon = Icon.GOLDEN_NOTES
            case Filter.RATING:
                icon = Icon.RATING
            case Filter.VIEWS:
                icon = Icon.VIEWS
            case Filter.YEAR:
                icon = Icon.CALENDAR
            case Filter.GENRE:
                icon = Icon.GENRE
            case Filter.CREATOR:
                icon = Icon.CREATOR
            case Filter.CUSTOM_DATA:
                icon = Icon.CUSTOM_DATA
            case _ as unreachable:
                assert_never(unreachable)
        return icon.icon()

    def is_checkable(self, has_checked_children: bool) -> bool:
        return self != Filter.SAVED and has_checked_children

    def is_editable(self) -> bool:
        return False

    def is_parent(self) -> bool:
        return True

    def is_accepted(self, _matches: set[str | int]) -> bool:
        return True

    def build_search(self, search: db.SearchBuilder) -> None:
        pass

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return True


class NodeItemData(TreeItemData):
    """Base implementation for a child of a filter item."""

    def decoration(self) -> QIcon | None:
        return None

    def is_checkable(self, has_checked_children: bool) -> bool:
        return True

    def is_editable(self) -> bool:
        return False

    def is_parent(self) -> bool:
        return False

    def is_accepted(self, _matches: set[str | int]) -> bool:
        return True

    def child_data(self) -> Iterable[TreeItemData]:
        return ()


T = TypeVar("T")


@attrs.define
class SongValueMatch(NodeItemData, Generic[T]):
    """str that can be matched against a specific attribute of a song."""

    val: T
    count: int

    def __str__(self) -> str:
        return f"{self.val} [{self.count}]"

    def search_attr(self, search: db.SearchBuilder) -> list[T]:
        raise NotImplementedError

    def build_search(self, search: db.SearchBuilder) -> None:
        self.search_attr(search).append(self.val)

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return self.val in self.search_attr(search)

    def is_accepted(self, matches: set[str | int]) -> bool:
        return self.val in matches


class SongArtistMatch(SongValueMatch):
    """str that can be matched against a song's artist."""

    def search_attr(self, search: db.SearchBuilder) -> list[str]:
        return search.artists


class SongTitleMatch(SongValueMatch):
    """str that can be matched against a song's title."""

    def search_attr(self, search: db.SearchBuilder) -> list[str]:
        return search.titles


class SongEditionMatch(SongValueMatch):
    """str that can be matched against a song's edition."""

    def search_attr(self, search: db.SearchBuilder) -> list[str]:
        return search.editions


class SongLanguageMatch(SongValueMatch):
    """str that can be matched against a song's language."""

    def search_attr(self, search: db.SearchBuilder) -> list[str]:
        return search.languages


class SongYearMatch(SongValueMatch):
    """str that can be matched against a song's year."""

    def search_attr(self, search: db.SearchBuilder) -> list[int]:
        return search.years


class SongGenreMatch(SongValueMatch):
    """str that can be matched against a song's genre."""

    def search_attr(self, search: db.SearchBuilder) -> list[str]:
        return search.genres


class SongCreatorMatch(SongValueMatch):
    """str that can be matched against a song's creator."""

    def search_attr(self, search: db.SearchBuilder) -> list[str]:
        return search.creators


class StatusVariant(NodeItemData, enum.Enum):
    """Variants of the status of a song."""

    NONE = db.DownloadStatus.NONE
    SYNCHRONIZED = db.DownloadStatus.SYNCHRONIZED
    OUTDATED = db.DownloadStatus.OUTDATED
    PENDING = db.DownloadStatus.PENDING
    DOWNLOADING = db.DownloadStatus.DOWNLOADING
    FAILED = db.DownloadStatus.FAILED

    def __str__(self) -> str:
        return "None" if self == StatusVariant.NONE else str(self.value)

    def build_search(self, search: db.SearchBuilder) -> None:
        search.statuses.append(self.value)

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return self.value in search.statuses


class RatingVariant(NodeItemData, enum.Enum):
    """Selectable variants for the song rating filter."""

    R_NONE = None
    R_1 = 1
    R_2 = 2
    R_3 = 3
    R_4 = 4
    R_5 = 5

    def __str__(self) -> str:
        if self == RatingVariant.R_NONE:
            return "None"
        return self.value * "★"

    def build_search(self, search: db.SearchBuilder) -> None:
        search.ratings.append(self.value or 0)

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return (self.value or 0) in search.ratings


class GoldenNotesVariant(NodeItemData, enum.Enum):
    """Selectable variants for the golden notes filter."""

    NO = False
    YES = True

    def __str__(self) -> str:
        match self:
            case GoldenNotesVariant.NO:
                return "No"
            case GoldenNotesVariant.YES:
                return "Yes"
            case _ as unreachable:
                assert_never(unreachable)

    def build_search(self, search: db.SearchBuilder) -> None:
        if search.golden_notes is None:
            search.golden_notes = self.value
        else:
            # Yes *and* No is the same as no filter
            search.golden_notes = None

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return self.value == search.golden_notes


class ViewsVariant(NodeItemData, enum.Enum):
    """Selectable variants for the views filter."""

    V_0 = (0, 100)
    V_100 = (100, 200)
    V_200 = (200, 300)
    V_300 = (300, 400)
    V_400 = (400, 500)
    V_500 = (500, None)

    def __str__(self) -> str:
        if self.value[1] is None:
            return f"{self.value[0]}+"
        return f"{self.value[0]} to {self.value[1] - 1}"

    def build_search(self, search: db.SearchBuilder) -> None:
        search.views.append(self.value)

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return self.value in search.views


@attrs.define
class SavedSearch(NodeItemData, db.SavedSearch):
    """A search saved by the user."""

    @classmethod
    def load_all(cls) -> Iterable[SavedSearch]:
        with db.transaction():
            searches = db.SavedSearch.load_saved_searches()
        return (cls(s.name, s.search, s.is_default, s.subscribed) for s in searches)

    def build_search(self, search: db.SearchBuilder) -> None:
        pass

    def __str__(self) -> str:
        return self.name

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return False

    def is_checkable(self, has_checked_children: bool) -> bool:
        return False

    def is_editable(self) -> bool:
        return True


@attrs.define
class CustomDataKeyMatch(NodeItemData):
    """Key-value pair that can be matched against a song's custom data."""

    key: str

    def __str__(self) -> str:
        return self.key

    def build_search(self, search: db.SearchBuilder) -> None:
        pass

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return True

    def is_accepted(self, matches: set[str | int]) -> bool:
        return True

    def child_data(self) -> Iterable[TreeItemData]:
        return (
            CustomDataMatch(self.key, v)
            for v in sorted(CustomData.value_options(self.key))
        )

    def is_parent(self) -> bool:
        return True

    def is_checkable(self, has_checked_children: bool) -> bool:
        return has_checked_children


@attrs.define
class CustomDataMatch(NodeItemData):
    """Key-value pair that can be matched against a song's custom data."""

    key: str
    value: str

    def __str__(self) -> str:
        return self.value

    def build_search(self, search: db.SearchBuilder) -> None:
        search.custom_data[self.key].append(self.value)

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return self.value in search.custom_data.get(self.key, [])

    def is_accepted(self, _matches: set[str | int]) -> bool:
        return True



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'PreviewWindow.ui'
##
## Created by: Qt User Interface Compiler version 6.9.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QGridLayout, QHBoxLayout,
    QLabel, QMainWindow, QSizePolicy, QSlider,
    QSpacerItem, QToolButton, QVBoxLayout, QWidget)
from usdb_syncer.gui.resources.qt import resources as resources_rc

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(1198, 600)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.horizontalLayout = QHBoxLayout(self.centralwidget)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.layout_main = QVBoxLayout()
        self.layout_main.setObjectName(u"layout_main")
        self.layout_buttons = QHBoxLayout()
        self.layout_buttons.setObjectName(u"layout_buttons")
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.layout_buttons.addItem(self.horizontalSpacer)

        self.button_to_start = QToolButton(self.centralwidget)
        self.button_to_start.setObjectName(u"button_to_start")
        icon = QIcon()
        icon.addFile(u":/icons/control-skip-180.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.button_to_start.setIcon(icon)
        self.button_to_start.setIconSize(QSize(20, 20))

        self.layout_buttons.addWidget(self.button_to_start)

        self.button_backward = QToolButton(self.centralwidget)
        self.button_backward.setObjectName(u"button_backward")
        icon1 = QIcon()
        icon1.addFile(u":/icons/control-double-180.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.button_backward.setIcon(icon1)
        self.button_backward.setIconSize(QSize(20, 20))

        self.layout_buttons.addWidget(self.button_backward)

        self.button_pause = QToolButton(self.centralwidget)
        self.button_pause.setObjectName(u"button_pause")
        icon2 = QIcon()
        icon2.addFile(u":/icons/control-pause.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.button_pause.setIcon(icon2)
        self.button_pause.setIconSize(QSize(20, 20))
        self.button_pause.setCheckable(True)

        self.layout_buttons.addWidget(self.button_pause)

        self.button_forward = QToolButton(self.centralwidget)
        self.button_forward.setObjectName(u"button_forward")
        icon3 = QIcon()
        icon3.addFile(u":/icons/control-double.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.button_forward.setIcon(icon3)
        self.button_forward.setIconSize(QSize(20, 20))

        self.layout_buttons.addWidget(self.button_forward)

        self.button_to_end = QToolButton(self.centralwidget)
        self.button_to_end.setObjectName(u"button_to_end")
        icon4 = QIcon()
        icon4.addFile(u":/icons/control-skip.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.button_to_end.setIcon(icon4)
        self.button_to_end.setIconSize(QSize(20, 20))

        self.layout_buttons.addWidget(self.button_to_end)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.layout_buttons.addItem(self.horizontalSpacer_2)


        self.layout_main.addLayout(self.layout_buttons)


        self.horizontalLayout.addLayout(self.layout_main)

        self.horizontalSpacer_3 = QSpacerItem(20, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer_3)

        self.layout_extra = QVBoxLayout()
        self.layout_extra.setObjectName(u"layout_extra")
        self.label_time = QLabel(self.centralwidget)
        self.label_time.setObjectName(u"label_time")
        font = QFont()
        font.setPointSize(12)
        font.setBold(True)
        self.label_time.setFont(font)
        self.label_time.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.layout_extra.addWidget(self.label_time)

        self.label_bpm = QLabel(self.centralwidget)
        self.label_bpm.setObjectName(u"label_bpm")

        self.layout_extra.addWidget(self.label_bpm)

        self.label_gap = QLabel(self.centralwidget)
        self.label_gap.setObjectName(u"label_gap")

        self.layout_extra.addWidget(self.label_gap)

        self.label_start = QLabel(self.centralwidget)
        self.label_start.setObjectName(u"label_start")

        self.layout_extra.addWidget(self.label_start)

        self.label_end = QLabel(self.centralwidget)
        self.label_end.setObjectName(u"label_end")

        self.layout_extra.addWidget(self.label_end)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.layout_extra.addItem(self.verticalSpacer)

        self.gridLayout = QGridLayout()
        self.gridLayout.setObjectName(u"gridLayout")
        self.label_ticks_volume = QLabel(self.centralwidget)
        self.label_ticks_volume.setObjectName(u"label_ticks_volume")
        self.label_ticks_volume.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.gridLayout.addWidget(self.label_ticks_volume, 2, 2, 1, 1)

        self.label_voice = QLabel(self.centralwidget)
        self.label_voice.setObjectName(u"label_voice")

        self.gridLayout.addWidget(self.label_voice, 0, 0, 1, 1)

        self.label = QLabel(self.centralwidget)
        self.label.setObjectName(u"label")

        self.gridLayout.addWidget(self.label, 1, 0, 1, 1)

        self.slider_source = QSlider(self.centralwidget)
        self.slider_source.setObjectName(u"slider_source")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.slider_source.sizePolicy().hasHeightForWidth())
        self.slider_source.setSizePolicy(sizePolicy)
        self.slider_source.setMaximum(10)
        self.slider_source.setSingleStep(1)
        self.slider_source.setPageStep(1)
        self.slider_source.setValue(10)
        self.slider_source.setOrientation(Qt.Orientation.Horizontal)

        self.gridLayout.addWidget(self.slider_source, 1, 1, 1, 1)

        self.label_source_volume = QLabel(self.centralwidget)
        self.label_source_volume.setObjectName(u"label_source_volume")
        self.label_source_volume.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.gridLayout.addWidget(self.label_source_volume, 1, 2, 1, 1)

        self.label_3 = QLabel(self.centralwidget)
        self.label_3.setObjectName(u"label_3")

        self.gridLayout.addWidget(self.label_3, 3, 0, 1, 1)

        self.label_pitch_volume = QLabel(self.centralwidget)
        self.label_pitch_volume.setObjectName(u"label_pitch_volume")
        self.label_pitch_volume.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.gridLayout.addWidget(self.label_pitch_volume, 3, 2, 1, 1)

        self.slider_ticks = QSlider(self.centralwidget)
        self.slider_ticks.setObjectName(u"slider_ticks")
        sizePolicy.setHeightForWidth(self.slider_ticks.sizePolicy().hasHeightForWidth())
        self.slider_ticks.setSizePolicy(sizePolicy)
        self.slider_ticks.setMaximum(10)
        self.slider_ticks.setSingleStep(1)
        self.slider_ticks.setPageStep(1)
        self.slider_ticks.setValue(10)
        self.slider_ticks.setOrientation(Qt.Orientation.Horizontal)

        self.gridLayout.addWidget(self.slider_ticks, 2, 1, 1, 1)

        self.label_2 = QLabel(self.centralwidget)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout.addWidget(self.label_2, 2, 0, 1, 1)

        self.comboBox_voice = QComboBox(self.centralwidget)
        self.comboBox_voice.setObjectName(u"comboBox_voice")

        self.gridLayout.addWidget(self.comboBox_voice, 0, 1, 1, 2)

        self.slider_pitch = QSlider(self.centralwidget)
        self.slider_pitch.setObjectName(u"slider_pitch")
        sizePolicy.setHeightForWidth(self.slider_pitch.sizePolicy().hasHeightForWidth())
        self.slider_pitch.setSizePolicy(sizePolicy)
        self.slider_pitch.setMaximum(10)
        self.slider_pitch.setPageStep(1)
        self.slider_pitch.setValue(0)
        self.slider_pitch.setOrientation(Qt.Orientation.Horizontal)

        self.gridLayout.addWidget(self.slider_pitch, 3, 1, 1, 1)


        self.layout_extra.addLayout(self.gridLayout)


        self.horizontalLayout.addLayout(self.layout_extra)

        MainWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"MainWindow", None))
        self.button_to_start.setText(QCoreApplication.translate("MainWindow", u"...", None))
        self.button_backward.setText(QCoreApplication.translate("MainWindow", u"...", None))
        self.button_pause.setText(QCoreApplication.translate("MainWindow", u"...", None))
        self.button_forward.setText(QCoreApplication.translate("MainWindow", u"...", None))
        self.button_to_end.setText(QCoreApplication.translate("MainWindow", u"...", None))
        self.label_time.setText(QCoreApplication.translate("MainWindow", u"00:00 / 02:30", None))
        self.label_bpm.setText(QCoreApplication.translate("MainWindow", u"#BPM: ", None))
        self.label_gap.setText(QCoreApplication.translate("MainWindow", u"#GAP: ", None))
        self.label_start.setText(QCoreApplication.translate("MainWindow", u"#START:", None))
        self.label_end.setText(QCoreApplication.translate("MainWindow", u"#END:", None))
        self.label_ticks_volume.setText(QCoreApplication.translate("MainWindow", u"100", None))
        self.label_voice.setText(QCoreApplication.translate("MainWindow", u"Voice:", None))
        self.label.setText(QCoreApplication.translate("MainWindow", u"Source:", None))
        self.label_source_volume.setText(QCoreApplication.translate("MainWindow", u"100", None))
        self.label_3.setText(QCoreApplication.translate("MainWindow", u"Pitch:", None))
        self.label_pitch_volume.setText(QCoreApplication.translate("MainWindow", u"0", None))
        self.label_2.setText(QCoreApplication.translate("MainWindow", u"Ticks:", None))
    # retranslateUi




"""Dialog to start and stop the webserver."""

from PySide6 import QtGui, QtWidgets

from usdb_syncer import webserver
from usdb_syncer.gui.forms.WebserverDialog import Ui_Dialog


class WebserverDialog(Ui_Dialog, QtWidgets.QDialog):
    """Dialog to start and stop the webserver."""

    def __init__(self, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent=parent)
        self.setupUi(self)
        self._update_ui()
        self.edit_title.setPlaceholderText(webserver.DEFAULT_TITLE)
        self.button_start.clicked.connect(self._start)
        self.button_stop.clicked.connect(self._stop)

    def _update_ui(self) -> None:
        running = webserver.is_running()
        self.edit_title.setEnabled(not running)
        self.box_port.setEnabled(not running)
        self.button_start.setEnabled(not running)
        self.button_stop.setEnabled(running)
        address = webserver.address()
        self.label_status.setText(
            f"The webserver is running on <a href='{address}'>{address}</a>."
            if running
            else "The webserver is not currently running."
        )
        pixmap = QtGui.QPixmap()
        if running:
            pixmap.loadFromData(webserver.get_qrcode(webserver.address()))
        self.label_qrcode.setPixmap(pixmap)

    def _start(self) -> None:
        webserver.start(title=self.edit_title.text(), port=self.box_port.value())
        self._update_ui()

    def _stop(self) -> None:
        webserver.stop()
        self._update_ui()



"""This file was generated using `write_resource_files.py`."""

from importlib import resources

NOTOSANS_BLACK_TTF = resources.files() / "NotoSans-Black.ttf"
NOTOSANS_BOLD_TTF = resources.files() / "NotoSans-Bold.ttf"
NOTOSANS_REGULAR_TTF = resources.files() / "NotoSans-Regular.ttf"



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'WebserverDialog.ui'
##
## Created by: Qt User Interface Compiler version 6.9.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QFormLayout, QHBoxLayout,
    QLabel, QLineEdit, QPushButton, QSizePolicy,
    QSpacerItem, QSpinBox, QVBoxLayout, QWidget)

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(348, 322)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.label_status = QLabel(Dialog)
        self.label_status.setObjectName(u"label_status")
        self.label_status.setOpenExternalLinks(True)

        self.verticalLayout.addWidget(self.label_status)

        self.formLayout = QFormLayout()
        self.formLayout.setObjectName(u"formLayout")
        self.label_2 = QLabel(Dialog)
        self.label_2.setObjectName(u"label_2")

        self.formLayout.setWidget(2, QFormLayout.ItemRole.LabelRole, self.label_2)

        self.box_port = QSpinBox(Dialog)
        self.box_port.setObjectName(u"box_port")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.box_port.sizePolicy().hasHeightForWidth())
        self.box_port.setSizePolicy(sizePolicy)
        self.box_port.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)
        self.box_port.setMaximum(65535)

        self.formLayout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.box_port)

        self.label = QLabel(Dialog)
        self.label.setObjectName(u"label")

        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.label)

        self.edit_title = QLineEdit(Dialog)
        self.edit_title.setObjectName(u"edit_title")

        self.formLayout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.edit_title)


        self.verticalLayout.addLayout(self.formLayout)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer_2)

        self.label_qrcode = QLabel(Dialog)
        self.label_qrcode.setObjectName(u"label_qrcode")
        self.label_qrcode.setMinimumSize(QSize(150, 150))
        self.label_qrcode.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.verticalLayout.addWidget(self.label_qrcode)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.button_start = QPushButton(Dialog)
        self.button_start.setObjectName(u"button_start")

        self.horizontalLayout.addWidget(self.button_start)

        self.button_stop = QPushButton(Dialog)
        self.button_stop.setObjectName(u"button_stop")

        self.horizontalLayout.addWidget(self.button_stop)


        self.verticalLayout.addLayout(self.horizontalLayout)


        self.retranslateUi(Dialog)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Webserver", None))
        self.label_status.setText(QCoreApplication.translate("Dialog", u"The webserver is not currently running.", None))
        self.label_2.setText(QCoreApplication.translate("Dialog", u"Port:", None))
        self.label.setText(QCoreApplication.translate("Dialog", u"Title:", None))
        self.label_qrcode.setText("")
        self.button_start.setText(QCoreApplication.translate("Dialog", u"Start", None))
        self.button_stop.setText(QCoreApplication.translate("Dialog", u"Stop", None))
    # retranslateUi




"""Contains a runnable song loader."""

from __future__ import annotations

import copy
import filecmp
import shutil
import tempfile
import time
from collections.abc import Iterable, Iterator
from itertools import islice
from pathlib import Path
from typing import ClassVar, assert_never

import attrs
import shiboken6
from PySide6 import QtCore

from usdb_syncer import (
    SongId,
    SyncMetaId,
    constants,
    db,
    download_options,
    errors,
    events,
    hooks,
    resource_dl,
    settings,
    usdb_scraper,
    utils,
)
from usdb_syncer.custom_data import CustomData
from usdb_syncer.discord import notify_discord
from usdb_syncer.logger import Logger, logger, song_logger
from usdb_syncer.postprocessing import write_audio_tags, write_video_tags
from usdb_syncer.resource_dl import ResourceDLError
from usdb_syncer.settings import FormatVersion
from usdb_syncer.song_txt import SongTxt
from usdb_syncer.sync_meta import ResourceFile, SyncMeta
from usdb_syncer.usdb_scraper import SongDetails
from usdb_syncer.usdb_song import DownloadStatus, UsdbSong
from usdb_syncer.utils import video_url_from_resource


class DownloadManager:
    """Manager for concurrent song downloads."""

    _jobs: ClassVar[dict[SongId, _SongLoader]] = {}
    _pause = False
    _pool: QtCore.QThreadPool | None = None

    @classmethod
    def download(cls, songs: Iterable[UsdbSong]) -> None:
        options = download_options.download_options()
        for song in songs:
            if song.song_id in cls._jobs:
                cls._jobs[song.song_id].logger.warning("Already downloading!")
                continue
            cls._jobs[song.song_id] = job = _SongLoader(song, options)
            job.pause = cls._pause
            cls._threadpool().start(job)

    @classmethod
    def abort(cls, songs: Iterable[SongId]) -> None:
        for song in songs:
            if (job := cls._jobs.get(song)) and shiboken6.isValid(job):
                if cls._threadpool().tryTake(job):
                    job.logger.info("Download aborted by user request.")
                    job.song.reset_status()
                    with db.transaction():
                        job.song.upsert()
                    events.SongChanged(job.song_id).post()
                    events.DownloadFinished(job.song_id).post()
                else:
                    job.abort = True

    @classmethod
    def set_pause(cls, pause: bool) -> None:
        cls._pause = pause
        for job in cls._jobs.values():
            job.pause = pause

    @classmethod
    def quit(cls) -> None:
        if cls._pool:
            logger.debug(f"Quitting {len(cls._jobs)} downloads.")
            for job in cls._jobs.values():
                job.abort = True
            cls._pool.waitForDone()

    @classmethod
    def _threadpool(cls) -> QtCore.QThreadPool:
        if cls._pool is None:
            cls._pool = QtCore.QThreadPool()
            if threads := settings.get_throttling_threads():
                cls._pool.setMaxThreadCount(threads)
            events.DownloadFinished.subscribe(cls._remove_job)
        return cls._pool

    @classmethod
    def _remove_job(cls, event: events.DownloadFinished) -> None:
        if event.song_id in cls._jobs:
            del cls._jobs[event.song_id]


@attrs.define(kw_only=True)
class _Locations:
    """Paths for downloading a song."""

    _current: Path | None
    # includes filename stem
    _target: Path
    _tempdir: Path

    @classmethod
    def new(
        cls, song: UsdbSong, options: download_options.Options, tempdir: Path
    ) -> _Locations:
        target = options.path_template.evaluate(song, options.song_dir)
        if (
            _current := song.sync_meta.path.parent if song.sync_meta else None
        ) and utils.path_matches_maybe_with_suffix(_current, target.parent):
            target = _current / target.name
        else:
            target = utils.next_unique_directory(target.parent) / target.name
        return cls(current=_current, target=target, tempdir=tempdir)  # pyright: ignore

    def current_path(self, file: str = "", ext: str = "") -> Path | None:
        """Path to file in the current download directory if it exists.
        The final path component is the generic name or the provided file, optionally
        with the provided extension joined with a '.' unless one is already present.
        """
        return self._path(self._current, file, ext) if self._current else None

    def temp_path(self, file: str = "", ext: str = "") -> Path:
        """Path to file in the temporary download directory.
        The final path component is the generic name or the provided file, optionally
        with the provided extension joined with a '.' unless one is already present.
        """
        return self._path(self._tempdir, file, ext)

    def target_path(self, file: str = "", ext: str = "") -> Path:
        """Path to file in the final download directory.
        The final path component is the generic name or the provided file, optionally
        with the provided extension joined with a '.' unless one is already present.
        """
        return self._path(self._target.parent, file, ext)

    def _path(self, parent: Path, file: str = "", ext: str = "") -> Path:
        name = file or self._target.name
        if ext:
            name = f"{name}{'' if '.' in ext else '.'}{ext}"
        return parent.joinpath(name)

    def filename(self, ext: str = "") -> str:
        if ext:
            return f"{self._target.name}{'' if '.' in ext else '.'}{ext}"
        return self._target.name

    def move_to_target_folder(self) -> None:
        """Rename the path of the song folder if it does not match the template, and
        ensure the target directory exists.
        """
        if self._current and self._current != self._target:
            self._target.parent.parent.mkdir(parents=True, exist_ok=True)
            self._current.rename(self._target.parent)
            self._current = self._target.parent
        else:
            self._target.parent.mkdir(parents=True, exist_ok=True)


@attrs.define
class _TempResourceFile:
    """Interim resource file in the temporary folder, or in the old folder if the
    resource is potentially kept.
    """

    old_fname: str | None = None
    new_fname: str | None = None
    resource: str | None = None

    def path_and_resource(
        self, locations: _Locations, temp: bool
    ) -> tuple[Path, str] | None:
        if (path := self.path(locations, temp=temp)) and self.resource:
            return (path, self.resource)
        return None

    def path(self, locations: _Locations, temp: bool) -> Path | None:
        if self.new_fname:
            if temp:
                return locations.temp_path(self.new_fname)
            return locations.target_path(self.new_fname)
        if self.old_fname:
            return locations.current_path(self.old_fname)
        return None

    def to_resource_file(
        self, locations: _Locations, temp: bool
    ) -> ResourceFile | None:
        if path_resource := self.path_and_resource(locations, temp=temp):
            return ResourceFile.new(*path_resource)
        return None


@attrs.define
class _TempResourceFiles:
    """Collection of all temporary resource files."""

    txt: _TempResourceFile = attrs.field(factory=_TempResourceFile)
    audio: _TempResourceFile = attrs.field(factory=_TempResourceFile)
    video: _TempResourceFile = attrs.field(factory=_TempResourceFile)
    cover: _TempResourceFile = attrs.field(factory=_TempResourceFile)
    background: _TempResourceFile = attrs.field(factory=_TempResourceFile)

    def __iter__(self) -> Iterator[_TempResourceFile]:
        return iter((self.txt, self.audio, self.video, self.cover, self.background))


@attrs.define
class _Context:
    """_Context for downloading media and creating a song folder."""

    # deep copy of the passed in song
    song: UsdbSong
    details: SongDetails
    options: download_options.Options
    txt: SongTxt
    locations: _Locations
    logger: Logger
    out: _TempResourceFiles = attrs.field(factory=_TempResourceFiles)

    def __attrs_post_init__(self) -> None:
        # reuse old resource files unless we acquire new ones later on
        if self.song.sync_meta and (current := self.locations.current_path()):
            for old, out in (
                (self.song.sync_meta.txt, self.out.txt),
                (self.song.sync_meta.audio, self.out.audio),
                (self.song.sync_meta.video, self.out.video),
                (self.song.sync_meta.cover, self.out.cover),
                (self.song.sync_meta.background, self.out.background),
            ):
                if old and old.is_in_sync(current.parent):
                    out.resource = old.resource
                    out.old_fname = old.fname

    @classmethod
    def new(
        cls,
        song: UsdbSong,
        options: download_options.Options,
        tempdir: Path,
        log: Logger,
    ) -> _Context:
        song = copy.deepcopy(song)
        details, txt = _get_usdb_data(song.song_id, options.txt_options, log)
        _update_song_with_usdb_data(song, details, txt)
        paths = _Locations.new(song, options, tempdir)
        if not song.sync_meta:
            song.sync_meta = SyncMeta.new(
                song.song_id, song.usdb_mtime, paths.target_path().parent, txt.meta_tags
            )
        return cls(song, details, options, txt, paths, log)

    def all_audio_resources(self) -> Iterator[str]:
        if self.txt.meta_tags.audio:
            yield self.txt.meta_tags.audio
        if not self.txt.meta_tags.video:
            self.logger.debug("No valid audio/video meta tag. Looking in comments.")
        yield from self.all_video_resources()

    def all_video_resources(self) -> Iterator[str]:
        if self.txt.meta_tags.video:
            yield self.txt.meta_tags.video
        yield from self.details.all_comment_videos()

    def background_url(self) -> str | None:
        url = None
        if self.txt.meta_tags.background:
            url = self.txt.meta_tags.background.source_url(self.logger)
            self.logger.debug(f"downloading background from #VIDEO params: {url}")
        return url


def _get_usdb_data(
    song_id: SongId, txt_options: download_options.TxtOptions | None, log: Logger
) -> tuple[SongDetails, SongTxt]:
    details = usdb_scraper.get_usdb_details(song_id)
    log.info(f"Found '{details.artist} - {details.title}' on USDB.")
    txt_str = usdb_scraper.get_notes(details.song_id, log)
    txt = SongTxt.parse(txt_str, log)
    txt.sanitize(txt_options)
    txt.headers.creator = txt.headers.creator or details.uploader or None
    return details, txt


def _update_song_with_usdb_data(
    song: UsdbSong, details: SongDetails, txt: SongTxt
) -> None:
    song.artist = details.artist
    song.title = details.title
    song.language = txt.headers.language or ""
    song.edition = txt.headers.edition or ""
    song.golden_notes = details.golden_notes
    song.rating = details.rating
    song.views = details.views
    if txt.headers.year and len(txt.headers.year) == 4 and txt.headers.year.isdigit():
        song.year = int(txt.headers.year)
    else:
        song.year = None
    song.genre = txt.headers.genre or ""
    song.creator = txt.headers.creator or ""


class _SongLoader(QtCore.QRunnable):
    """Runnable to create a complete song folder."""

    abort = False
    pause = False

    def __init__(self, song: UsdbSong, options: download_options.Options) -> None:
        super().__init__()
        self.song = song
        self.song_id = song.song_id
        self.options = options
        self.logger = song_logger(self.song_id)

    def run(self) -> None:
        with db.managed_connection(utils.AppPaths.db):
            try:
                self.song = self._run_inner()
            except errors.AbortError:
                self.logger.info("Download aborted by user request.")
                self.song.reset_status()
            except errors.UsdbLoginError:
                self.logger.error("Aborted; download requires login.")  # noqa: TRY400
                self.song.status = DownloadStatus.FAILED
            except errors.UsdbNotFoundError:
                self.logger.error("Song has been deleted from USDB.")  # noqa: TRY400
                with db.transaction():
                    self.song.delete()
                if meta := self.song.sync_meta:
                    path = meta.path.parent
                    self.logger.info(f"Trashing local song {path}")
                    utils.trash_or_delete_path(path)
                events.SongDeleted(self.song_id).post()
                events.DownloadFinished(self.song_id).post()
                return
            except Exception:
                self.logger.exception(
                    "Failed to finish download due to an unexpected error. "
                    "See debug log for more information."
                )
                self.song.status = DownloadStatus.FAILED
            else:
                self.song.status = DownloadStatus.SYNCHRONIZED
                self.logger.info("All done!")
            with db.transaction():
                self.song.upsert()
        events.SongChanged(self.song_id).post()
        events.DownloadFinished(self.song_id).post()

    def _run_inner(self) -> UsdbSong:
        self._check_flags()
        self.song.status = DownloadStatus.DOWNLOADING
        with db.transaction():
            self.song.upsert()
        events.SongChanged(self.song_id).post()
        with tempfile.TemporaryDirectory() as tempdir:
            ctx = _Context.new(self.song, self.options, Path(tempdir), self.logger)
            for job in (
                _maybe_download_audio,
                _maybe_download_video,
                _maybe_download_cover,
                _maybe_download_background,
                _maybe_write_audio_tags,
                _maybe_write_video_tags,
            ):
                self._check_flags()
                job(ctx)

            # last chance to abort before irreversible changes
            self._check_flags()
            _cleanup_existing_resources(ctx)
            # only here so filenames in header are up-to-date
            _maybe_write_txt(ctx)
            ctx.locations.move_to_target_folder()
            _persist_tempfiles(ctx)
        _write_sync_meta(ctx)
        hooks.SongLoaderDidFinish.call(ctx.song)
        return ctx.song

    def _check_flags(self) -> None:
        if self.abort:
            raise errors.AbortError
        if self.pause:
            while self.pause:
                time.sleep(0.5)
                if self.abort:
                    raise errors.AbortError


def _maybe_download_audio(ctx: _Context) -> None:
    if not (options := ctx.options.audio_options):
        return
    for resource in islice(ctx.all_audio_resources(), 10):
        if ctx.out.audio.resource == resource:
            ctx.logger.info("Audio resource is unchanged.")
            return
        dl_result = resource_dl.download_audio(
            resource,
            options,
            ctx.options.browser,
            ctx.locations.temp_path(),
            ctx.logger,
        )
        if ext := dl_result.extension:
            ctx.out.audio.resource = resource
            ctx.out.audio.new_fname = ctx.locations.filename(ext=ext)
            ctx.logger.info("Success! Downloaded audio.")
            return
        if dl_result.error in {
            resource_dl.ResourceDLError.RESOURCE_INVALID,
            resource_dl.ResourceDLError.RESOURCE_UNSUPPORTED,
            resource_dl.ResourceDLError.RESOURCE_UNAVAILABLE,
            resource_dl.ResourceDLError.RESOURCE_PARSE_ERROR,
        }:
            if ctx.options.notify_discord and (
                url := video_url_from_resource(resource)
            ):
                notify_discord(
                    ctx.song.song_id, url, "Audio", dl_result.error.value, logger
                )
    keep = " Keeping last resource." if ctx.out.audio.resource else ""
    song_len = ctx.txt.minimum_song_length()
    ctx.logger.error(f"Failed to download audio (song duration > {song_len})!{keep}")


def _maybe_download_video(ctx: _Context) -> None:
    if not (options := ctx.options.video_options) or ctx.txt.meta_tags.is_audio_only():
        return
    for resource in islice(ctx.all_video_resources(), 10):
        if ctx.out.video.resource == resource:
            ctx.logger.info("Video resource is unchanged.")
            return
        dl_result = resource_dl.download_video(
            resource,
            options,
            ctx.options.browser,
            ctx.locations.temp_path(),
            ctx.logger,
        )
        if ext := dl_result.extension:
            ctx.out.video.resource = resource
            ctx.out.video.new_fname = ctx.locations.filename(ext=ext)
            ctx.logger.info("Success! Downloaded video.")
            return
        if dl_result.error in {
            resource_dl.ResourceDLError.RESOURCE_INVALID,
            resource_dl.ResourceDLError.RESOURCE_UNSUPPORTED,
            resource_dl.ResourceDLError.RESOURCE_UNAVAILABLE,
            resource_dl.ResourceDLError.RESOURCE_PARSE_ERROR,
        }:
            if ctx.options.notify_discord and (
                url := video_url_from_resource(resource)
            ):
                notify_discord(
                    ctx.song.song_id, url, "Video", dl_result.error.value, logger
                )
    keep = " Keeping last resource." if ctx.out.video.resource else ""
    ctx.logger.error(f"Failed to download video!{keep}")


def _maybe_download_cover(ctx: _Context) -> None:
    if not ctx.options.cover:
        return
    if ctx.txt.meta_tags.cover is None and ctx.details.cover_url is None:
        ctx.logger.warning("No cover resource found.")
        return
    if cover := ctx.txt.meta_tags.cover:
        url = cover.source_url(ctx.logger)
        if _download_cover_url(ctx, url):
            return
        if ctx.options.notify_discord:
            notify_discord(
                ctx.song.song_id,
                url,
                "Cover",
                ResourceDLError.RESOURCE_UNAVAILABLE.value,
                ctx.logger,
            )
    if ctx.details.cover_url:
        ctx.logger.warning("Falling back to small USDB cover.")
        if _download_cover_url(ctx, ctx.details.cover_url, process=False):
            return
    keep = " Keeping last resource." if ctx.out.cover.resource else ""
    ctx.logger.error(f"Failed to download cover!{keep}")


def _download_cover_url(ctx: _Context, url: str, process: bool = True) -> bool:
    """True if download was successful (or is unnecessary)."""
    assert ctx.options.cover
    if ctx.out.cover.resource == url:
        if sync_meta := ctx.song.sync_meta:
            if sync_meta.meta_tags.cover == ctx.txt.meta_tags.cover:
                ctx.logger.info(
                    "Cover resource and postprocessing parameters are unchanged, "
                    "skipping."
                )
                return True
            ctx.logger.info(
                "Cover postprocessing parameters have changed, redownloading."
            )
    if path := resource_dl.download_and_process_image(
        url=url,
        target_stem=ctx.locations.temp_path(),
        meta_tags=ctx.txt.meta_tags.cover,
        details=ctx.details,
        kind=resource_dl.ImageKind.COVER,
        max_width=ctx.options.cover.max_size,
        process=process,
    ):
        ctx.out.cover.resource = url
        ctx.out.cover.new_fname = path.name
        ctx.logger.info("Success! Downloaded cover.")
        return True
    return False


def _maybe_download_background(ctx: _Context) -> None:
    if not (options := ctx.options.background_options):
        return
    if not options.download_background(bool(ctx.out.video.resource)):
        return
    if not (url := ctx.background_url()):
        ctx.logger.warning("No background resource found.")
        return
    if ctx.out.background.resource == url:
        if sync_meta := ctx.song.sync_meta:
            if sync_meta.meta_tags.background == ctx.txt.meta_tags.background:
                ctx.logger.info(
                    "Background resource and postprocessing parameters are unchanged, "
                    "skipping."
                )
                return
            ctx.logger.info(
                "Background postprocessing parameters have changed, redownloading."
            )
    if path := resource_dl.download_and_process_image(
        url=url,
        target_stem=ctx.locations.temp_path(),
        meta_tags=ctx.txt.meta_tags.background,
        details=ctx.details,
        kind=resource_dl.ImageKind.BACKGROUND,
        max_width=None,
    ):
        ctx.out.background.resource = url
        ctx.out.background.new_fname = path.name
        ctx.logger.info("Success! Downloaded background.")
    else:
        if ctx.options.notify_discord:
            notify_discord(
                ctx.song.song_id,
                url,
                "Background",
                ResourceDLError.RESOURCE_UNAVAILABLE.value,
                ctx.logger,
            )
        keep = " Keeping last resource." if ctx.out.cover.resource else ""
        ctx.logger.error(f"Failed to download background!{keep}")


def _maybe_write_txt(ctx: _Context) -> None:
    if not (options := ctx.options.txt_options):
        return
    _write_headers(ctx)
    path = ctx.locations.temp_path(ext="txt")
    ctx.txt.write_to_file(path, options.encoding.value, options.newline.value)
    if (
        ctx.out.txt.old_fname
        and (old_path := ctx.locations.current_path(ctx.out.txt.old_fname))
        and filecmp.cmp(path, old_path, shallow=False)
    ):
        ctx.logger.info("Song txt is unchanged.")
    else:
        ctx.out.txt.new_fname = path.name
        ctx.out.txt.resource = ctx.song.song_id.usdb_gettxt_url()
        ctx.logger.info("Success! Created song txt.")


def _write_headers(ctx: _Context) -> None:
    version = (
        ctx.options.txt_options.format_version
        if ctx.options and ctx.options.txt_options
        else FormatVersion.V1_0_0
    )

    if version >= FormatVersion.V1_1_0:
        ctx.txt.headers.providedby = constants.Usdb.BASE_URL

    if path := ctx.out.audio.path(ctx.locations, temp=True):
        _set_audio_headers(ctx, version, path)

    if path := ctx.out.video.path(ctx.locations, temp=True):
        _set_video_headers(ctx, version, path)

    if path := ctx.out.cover.path(ctx.locations, temp=True):
        _set_cover_headers(ctx, version, path)

    if path := ctx.out.background.path(ctx.locations, temp=True):
        _set_background_headers(ctx, version, path)


def _set_audio_headers(ctx: _Context, version: FormatVersion, path: Path) -> None:
    match version:
        case FormatVersion.V1_0_0:
            ctx.txt.headers.mp3 = path.name
        case FormatVersion.V1_1_0:
            # write both #MP3 and #AUDIO to maximize compatibility
            ctx.txt.headers.mp3 = path.name
            ctx.txt.headers.audio = path.name
        case FormatVersion.V1_2_0:
            ctx.txt.headers.audio = path.name
            if resource := ctx.txt.meta_tags.audio or ctx.txt.meta_tags.video:
                ctx.txt.headers.audiourl = video_url_from_resource(resource)
        case _ as unreachable:
            assert_never(unreachable)


def _set_video_headers(ctx: _Context, version: FormatVersion, path: Path) -> None:
    ctx.txt.headers.video = path.name
    if version >= FormatVersion.V1_2_0 and (resource := ctx.txt.meta_tags.video):
        ctx.txt.headers.videourl = video_url_from_resource(resource)


def _set_cover_headers(ctx: _Context, version: FormatVersion, path: Path) -> None:
    ctx.txt.headers.cover = path.name
    if (
        version >= FormatVersion.V1_2_0
        and ctx.txt.meta_tags.cover
        and (url := ctx.txt.meta_tags.cover.source_url(ctx.logger))
    ):
        ctx.txt.headers.coverurl = url


def _set_background_headers(ctx: _Context, version: FormatVersion, path: Path) -> None:
    ctx.txt.headers.background = path.name
    if (
        version >= FormatVersion.V1_2_0
        and ctx.txt.meta_tags.background
        and (url := ctx.txt.meta_tags.background.source_url(ctx.logger))
    ):
        ctx.txt.headers.backgroundurl = url


def _maybe_write_audio_tags(ctx: _Context) -> None:
    if not (options := ctx.options.audio_options):
        return
    if not (
        audio_path_resource := ctx.out.audio.path_and_resource(ctx.locations, temp=True)
    ):
        return
    cover_path_resource = ctx.out.cover.path_and_resource(ctx.locations, temp=True)
    background_path_resource = ctx.out.background.path_and_resource(
        ctx.locations, temp=True
    )
    write_audio_tags(
        txt=ctx.txt,
        options=options,
        audio=audio_path_resource,
        cover=cover_path_resource,
        background=background_path_resource,
        logger=ctx.logger,
    )


def _maybe_write_video_tags(ctx: _Context) -> None:
    if not (options := ctx.options.video_options):
        return
    if not (
        video_path_resource := ctx.out.video.path_and_resource(ctx.locations, temp=True)
    ):
        return
    cover_path_resource = ctx.out.cover.path_and_resource(ctx.locations, temp=True)
    background_path_resource = ctx.out.background.path_and_resource(
        ctx.locations, temp=True
    )
    write_video_tags(
        txt=ctx.txt,
        options=options,
        video=video_path_resource,
        cover=cover_path_resource,
        background=background_path_resource,
        logger=ctx.logger,
    )


def _cleanup_existing_resources(ctx: _Context) -> None:
    """Delete resources that are either out of sync or will be replaced with a new one,
    and ensure kept ones are correctly named.
    """
    if not ctx.song.sync_meta:
        return
    for (old, _), out in zip(
        ctx.song.sync_meta.all_resource_files(), ctx.out, strict=False
    ):
        if not (old and (old_path := ctx.locations.current_path(file=old.fname))):
            continue
        if not out.old_fname:
            # out of sync
            if old_path.exists():
                utils.trash_or_delete_path(old_path)
                ctx.logger.debug(f"Trashed untracked file: '{old_path}'.")
        elif out.new_fname:
            utils.trash_or_delete_path(old_path)
            ctx.logger.debug(f"Trashed existing file: '{old_path}'.")
        else:
            target = ctx.locations.filename(ext=utils.resource_file_ending(old.fname))
            if out.old_fname != target:
                # no new file; keep existing one, but ensure correct name
                path = old_path.with_name(target)
                old_path.rename(path)
                out.old_fname = target


def _persist_tempfiles(ctx: _Context) -> None:
    for temp_file in ctx.out:
        if temp_file.new_fname and (
            temp_path := temp_file.path(ctx.locations, temp=True)
        ):
            target = ctx.locations.target_path(temp_path.name)
            if target.exists():
                utils.trash_or_delete_path(target)
                ctx.logger.debug(f"Trashed existing file: '{target}'.")
            shutil.move(temp_path, target)


def _write_sync_meta(ctx: _Context) -> None:
    old = ctx.song.sync_meta
    sync_meta_id = old.sync_meta_id if old else SyncMetaId.new()
    ctx.song.sync_meta = SyncMeta(
        sync_meta_id=sync_meta_id,
        song_id=ctx.song.song_id,
        usdb_mtime=ctx.song.usdb_mtime,
        path=ctx.locations.target_path(file=sync_meta_id.to_filename()),
        mtime=0,
        meta_tags=ctx.txt.meta_tags,
        pinned=old.pinned if old else False,
        custom_data=CustomData(old.custom_data.inner() if old else None),
    )
    ctx.song.sync_meta.txt = ctx.out.txt.to_resource_file(ctx.locations, temp=False)
    ctx.song.sync_meta.audio = ctx.out.audio.to_resource_file(ctx.locations, temp=False)
    ctx.song.sync_meta.video = ctx.out.video.to_resource_file(ctx.locations, temp=False)
    ctx.song.sync_meta.cover = ctx.out.cover.to_resource_file(ctx.locations, temp=False)
    ctx.song.sync_meta.background = ctx.out.background.to_resource_file(
        ctx.locations, temp=False
    )
    ctx.song.sync_meta.synchronize_to_file()



"""Signals other components can notify and subscribe to."""

import attrs

from usdb_syncer import db, events
from usdb_syncer.gui import theme

# search


@attrs.define(slots=False)
class TreeFilterChanged(events.SubscriptableEvent):
    """Sent when a tree filter row has been selected or deselected."""

    search: db.SearchBuilder


@attrs.define(slots=False)
class TextFilterChanged(events.SubscriptableEvent):
    """Sent when the free text search has been changed."""

    search: str


@attrs.define(slots=False)
class SearchOrderChanged(events.SubscriptableEvent):
    """Sent when the search order has been changed or reversed."""

    order: db.SongOrder
    descending: bool


@attrs.define(slots=False)
class SavedSearchRestored(events.SubscriptableEvent):
    """Sent when the a save search is set."""

    search: db.SearchBuilder


# UI


@attrs.define(slots=False)
class ThemeChanged(events.SubscriptableEvent):
    """Sent when a new theme has been applied."""

    theme: theme.Theme
